// $Id: ASC.h 1316 2010-11-26 20:51:37Z mike $
// Created:   25.03.2006 lohmann

/**
 * \file ASC.h
 * \brief Definition of hardware access classes for the TC1796 ASC 0/1 serial interfaces 
 * 
 */

/**
 * @defgroup halstd Standard HAL Abstractions (mandatory, available on every plattform)
 * @defgroup halext Extended HAL Abstractions (optional, available on some plattforms)
 * @defgroup halspecific Plattform-specific HAL Extensions
 */
 
#ifndef HW_DEV_ASC_H_
#define HW_DEV_ASC_H_

#include "stdinc.h"
#include "hw/triboard1796.h"
#include "hw/irq/ASC.h"

//#if __TOOL_VERSION__ >= 210
#include <tc1796b/asc0-struct.h>
typedef ASC0_t ASC_t;
#include <tc1796b/asc1-struct.h>
//#else
//#include <tc1796b/asc-struct.h>
//#endif

#include <tc1796b/asc0/addr.h>

namespace hw { namespace dev {

/**
 * \class ASCBase
 * \brief TC1796 common base for the ASC0 and ASC1 serial interfaces.
 * 
 * The two serial interfaces of the TC1796 are accessed by (almost) identical
 * register sets, so most functionality can be implemented in this baseclass
 * that contains a pointer to the actual register set to use
 */
class ASCBase {
  protected:
	ASC_t* asc_;	// ASCx module base pointer

  public:
  	
  	/**
 	 * @brief Ready to send returns true, if transmit buffer is empty
	 */
	Bool RTS() const {
		return asc_->TBSRC.bits.SRR;
	}
	//! Sets/clears the RTS bit
	void RTS( Bool b ) {
		if( b ) asc_->TBSRC.bits.SETR = 1; else  asc_->TBSRC.bits.CLRR = 1 ;
	}
	
	//! Ready to read returns true, if a receive buffer contains a character
	Bool RTR() const {
		return asc_->RSRC.bits.SRR;
	}
	//! Sets/clears the RTR bit
	void RTR( Bool b ) {
		if( b ) asc_->RSRC.bits.SETR = 1; else  asc_->RSRC.bits.CLRR = 1 ;
	}
	
	//! Writes a character to the transmit buffer
	void send( UInt8 c) {
		asc_->TBUF.bits.TD_VALUE = c;	
	}
	
	//! Waits until RTS() == true and transmits c
	void putc_( UInt8 c );
	
	//! Reads a character from the receive buffer
	UInt8 receive() {
		return (UInt8) asc_->RBUF.bits.RD_VALUE;
	}
	
	//! Waits until RTR() == true and returns the read character
	UInt8 getc_();

	//! Sets the serial interface parameters.
	//
	//! @note On TC1796, both interfaces ASC0 and ASC1 are driven
	//! by the same baud rate clock timer.
	void setMode( UInt32 baud );	
};

/**
 * \class ASC0
 * \brief TC1796 serial interface 0 implementation.
 * 
 */
class ASC0 : public ASCBase {
	static ASC0 theASC0_;
	void init();
  public:
	
	static ASC0& Inst() {
		return theASC0_;
	}

	//! Class implementing our receive IRQ 
	typedef hw::irq::ASC0_REC IRQ_REC;

	//! Class implementing our transmit buffer IRQ 
	typedef hw::irq::ASC0_TBUF IRQ_TBUF;
	
	//! Class implementing our transmit IRQ 
	typedef hw::irq::ASC0_TRA IRQ_TRA;

	//! Class implementing our error IRQ 
	typedef hw::irq::ASC0_ERR IRQ_ERR;
};

/**
 * \class ASC1
 * \brief TC1796 serial interface 1 implementation
 *
 */

class ASC1 : public ASCBase {
	static ASC1 theASC1_;
	void init();
  public:
	
	static ASC1& Inst() {
		return theASC1_;
	}
	//! Class implementing our receive IRQ 
	typedef hw::irq::ASC1_REC IRQ_REC;

	//! Class implementing our transmit buffer IRQ 
	typedef hw::irq::ASC1_TBUF IRQ_TBUF;
	
	//! Class implementing our transmit IRQ 
	typedef hw::irq::ASC1_TRA IRQ_TRA;

	//! Class implementing our error IRQ 
	typedef hw::irq::ASC1_ERR IRQ_ERR;
};




}} // namespace hw::dev
#endif /*HW_DEV_ASC_H_*/
