// $Id: STM.h 1276 2010-09-14 10:27:54Z mike $
// Created:   Apr 11, 2006 lohmann

/**
 * @file STM.h
 * @brief Provides hardware access classes for the TriCore System Time Module (STM)
 *
 */

#ifndef __HW_DEV_STM_H__
#define __HW_DEV_STM_H__

#include "stdinc.h"
#include "hw/dev/CGU.h"
#include "hw/irq/STM.h"

// Include trigcc hw-support headers
#include <tc1796b/stm.h>

namespace hw { namespace dev {

/**
 * @class STM
 * @brief TC1796 system timer/counter access class
 * 
 * The TriCore STM implements a 56 bit incrementing counter driven at a maximum of 75 MHz.
 * (overflow after ~30 years). Two independend 32bit compare-and-match registers
 * are supported to trigger IRQs on match, thus providing two independend repetetive timer.
 *
 * Current counter values are available through 32 bit registers, the high 24 bits are always
 * latched (STM_CAP) if one of the 32 bit registers (STM_TIM0-STM_TIM5) is read. These registers
 * provide different excerpts from the 56 bit counter value:
 * @verbatim
 *            55      47      39      31      23      15      07     00
 *           +-------+-------+-------+-------+-------+-------+-------+
 *           |       |       |   The 56 Bit System Timer     |       |
 *           +-------+-------+-------+-------+---+---+---+---+---+---+
 *           |                       |           |   |   |   |   |   |
 *           |                       |           |   |   |   |   |   |
 *   +-------+-------+-------+-------+           |   |   |   |   |   |
 *   | 0x00  | 24 high bits (latched)| STM_CAP   |   |   |   |   |   |
 *   +-------+-------+-------+-------+           |   |   |   |   |   |
 *           |                       |           |   |   |   |   |   |
 *           |                       |           |   |   |   |   |   |
 *   +-------+-------+-------+-------+           |   |   |   |   |   |
 *   | 0x00  | 24 bit excerpt|       | STM_TIM6  |   |   |   |   |   |
 *   +-------+-------+-------+-------+           |   |   |   |   |   |
 *                51                             20  |   |   |   |   |
 *               +-------+-------+-------+-------+   |   |   |   |   |
 *   STM_TIM5    |       | 32 bit excerpt|       |   |   |   |   |   |
 *               +-------+-------+-------+-------+   |   |   |   |   |
 *                    47                             16  |   |   |   |
 *                   +-------+-------+-------+-------+   |   |   |   |
 *   STM_TIM4        |       | 32 bit excerpt|       |   |   |   |   |
 *                   +-------+-------+-------+-------+   |   |   |   |
 *                        43                             12  |   |   |
 *                       +-------+-------+-------+-------+   |   |   |
 *   STM_TIM3            |       | 32 bit excerpt|       |   |   |   |
 *                       +-------+-------+-------+-------+   |   |   |
 *                            39                             08  |   |
 *                           +-------+-------+-------+-------+   |   |
 *   STM_TIM2                |       | 32 bit excerpt|       |   |   |
 *                           +-------+-------+-------+-------+   |   |
 *                                35                             04  |
 *                               +-------+-------+-------+-------+   |
 *   STM_TIM1                    |       | 32 bit excerpt|       |   |
 *                               +-------+-------+-------+-------+   |
 *                                    31                             00
 *                                   +-------+-------+-------+-------+ 
 *   STM_TIM0                        |       | 32 bit excerpt|       | 
 *                                   +-------+-------+-------+-------+ 
 * @endverbatim
 * 
 * The STM always starts counting after a reset. 
 * 
 */
class STM {
	static STM theSTM_;
	//! Performs start-time initializations. Only called by advice.
	void init();
  protected:
    STM(){}
  public:	
	//! Returns the one-and-only STM instance
	static STM& Inst() {
		return theSTM_;
	}
		
	//! returns the lower 32 bits of the current counter value
	static UInt32 getLowValue() {
		return STM_TIM0.reg;
	}

	//! returns the upper 24 bits of the current counter value.
	static UInt32 getHighValue() {
		return STM_TIM6.reg;
	}

	//! returns the upper 24 bits of counter value latched on the last call to getLowValue()
	static UInt32 getLatchedHighValue() {
		return STM_CAP.reg;
	}
	
	//! returns all 56 bits of the current counter value as a 64bit unsigned integer
	static UInt64 value() {
		UInt64 res;
		
		// What we want to achieve is the following:
		//     res = (Value) getLowValue() | getHighValue() << 32;
		//
		// Unfortunately the compiler does not optimze this. It does not
		// notice that 'shift' and 'or' are actually not necessary, as
		// we shift by 32 which is a full register size and we therefore can simply 'move' 
		// the high 32 bits into the upper register of the register pair housing the 64 bit value
		asm volatile (	"ld.w %L0, STM_TIM0\n" 
						"ld.w %H0, STM_CAP\n"
						: "=d" (res) );
		return res;
	}
	
	enum DOXY_SYMBOL(Consts) {
		SYS_DIV				= cfHW_STM_DIV,				//!< divider relative to system frequency (see @ref hw::dev::CGU::Consts)
		FREQ				= CGU::SYSFREQ / SYS_DIV	//!< resulting counter clock frequency
	};

	//! returns the counter clock frequency
	static UInt32 freq() {
		return CGU::Inst().getSysClockFreq() / sysDivider();
	}

	//! returns the system clock (SYA) to counter clock (STM) divider
	static UInt32 sysDivider() {
		return SYS_DIV;
	}
	
	/**
	 * @brief Calculates timer cmp, size and start values from an intended period length.
	 *
	 * @param period is the intended period lenght in resulution units (getResUnit())
	 * @param cmp is the output compare value to be set for the CMP0() CMP1() register
	 * @param size is the output significant compare word size to be set for CMCON().MSIZEx
	 * @param start is the first significant bit from the 56 bit counter value to be used for the 
	 *        comparison. To be set into CMCON().MSIZEx
	 * @return is the error in resoulition units the thereby resulting period is truncated compared
	 * to \a period.
	 */	
	static UInt32 periodToCompareValues( UInt64 period, UInt32& cmp, unsigned& size, unsigned& start ); 	

	//! Get value of STM_TIMx registers
	template< int N >
	static UInt32 volatile TIM() {
		STATIC_ASSERT( ( (N >= 0) && (N <= 6) ), STM_TIMx_OUT_OF_RANGE);
		return *( ((UInt32*) STM_TIM0_ADDR) + N );	
	}
	
	//! Get/set compare-and-match register 0.
	static UInt32& CMP0() {
		return *( (UInt32*) STM_CMP0_ADDR);
	}
	
	//! Get/set compare-and-match register 1.
	static UInt32& CMP1() {
		return *( (UInt32*) STM_CMP1_ADDR);
	}
	
	//! Get/set compare-and-match control register.
	static STM_CMCON_t& CMCON() {
		return STM_CMCON;
	}
	
	//! Get/set interrupt control register.
	static STM_ICR_t& ICR() {
		return STM_ICR;
	}

	//! Set service set/reset register. Used to set/reset the internal interrupt request bits.
	static STM_ISRR_t& ISRR() {
		return STM_ISRR;
	}
	
	//! The TriCore SRN 0, which triggers the configured internal interrupts to the CPU
	typedef hw::irq::STM_SRC0 IRQ_SRC0;
	
	//! The TriCore SRN 1 which triggers the configured internal interrupts to the CPU
	typedef hw::irq::STM_SRC1 IRQ_SRC1;
};



}} // namespace hw::dev


#endif // __HW_DEV_STM_H__
