#ifndef _HW_DEV_COM_H_
#define _HW_DEV_COM_H_

#include "stdinc.h"

#include "IOPort.h"

namespace hw {
namespace dev {

template<UInt16 T_PORTADDR>
class COM {

public:
	enum Parity {
		NO,
		ODD,
		EVEN
	};

	enum StopBits {
		ONE,
		TWO
	};

	enum DataBits {
		FIVE,
		SIX,
		SEVEN,
		EIGHT
	};

private:

	struct {
		IOPort lcr;
		IOPort buffer;
		IOPort ier;
		IOPort iir;
		IOPort lsr;
	} regs_;

private:
	void setDLAB(bool on) {
		regs_.lcr.outb((regs_.lcr.inb() & ~0x80) | (on << 7));
	}

public:
	static const UInt16 PORTBASE = T_PORTADDR;

	void init() {
		regs_.lcr.setAddress(PORTBASE + 3);
		regs_.buffer.setAddress(PORTBASE + 0);
		regs_.ier.setAddress(PORTBASE + 1);
		regs_.iir.setAddress(PORTBASE + 2);
		regs_.lsr.setAddress(PORTBASE + 5);

		// set defaults (57600n1)
		setMode();

		enableInt();
	}

	void setBaudRate(UInt32 baudrate) {
		UInt32 quot = 115200 / baudrate;

		setDLAB(true);
		regs_.ier.outb(quot >> 8);
		regs_.buffer.outb(quot & 0xFF);
		setDLAB(false);
	}

	void setParity(Parity parity) {
		UInt8 lcr = regs_.lcr.inb() & ~(7 << 3);

		switch (parity) {
		case NO:
			break;

		case ODD:
			lcr |= 1 << 3;
			break;

		case EVEN:
			lcr |= 3 << 3;
			break;
		}

		regs_.lcr.outb(lcr);
	}

	void setStopBits(StopBits stopbits) {
		UInt8 lcr = regs_.lcr.inb() & ~(1 << 2);

		switch (stopbits) {
		case ONE:
			break;

		case TWO:
			lcr |= 1 << 2;
			break;
		}

		regs_.lcr.outb(lcr);
	}

	void setDataBits(DataBits databits) {
		UInt8 lcr = regs_.lcr.inb() & ~(3 << 0);

		switch (databits) {
		case FIVE:
			break;

		case SIX:
			lcr |= 1;
			break;

		case SEVEN:
			lcr |= 2;
			break;

		case EIGHT:
			lcr |= 3;
			break;
		}

		regs_.lcr.outb(lcr);
	}

	void enableInt(bool enable = true) {
		if (enable)
			regs_.ier.outb(0x3);
		else
			regs_.ier.outb(0x0);
	}

	void setMode(unsigned baudrate = 57600, Parity parity = NO, StopBits stopbits = ONE, DataBits databits = EIGHT) {
		setBaudRate(baudrate);
		setParity(parity);
		setStopBits(stopbits);
		setDataBits(databits);
	}

	void putc(char c) {
		while (! (regs_.lsr.inb() & ( 1 << 5 ))) {
			// Transmitter voll
		}

		regs_.buffer.outb(c);
	}

	char getc() {
		while (! (regs_.lsr.inb() & ( 1 << 0 ))) {
			// Receiver leer
		}

		return regs_.buffer.inb();
	}

	UInt8 getIIR() {
		return regs_.iir.inb();
	}
};

}
}

#endif
