#include "E1000.h"

#include <stdio.h>

namespace hw {
namespace dev {

E1000 E1000::inst_;

void E1000::init_regs() {
	// Check, if E1000 is present
	const PCIConfigSpace* cs = PCI::Inst().getConfigSpaceEntry(DEVICE, VENDOR);
	if (!cs)
		return;

	vector = cs->interrupt_line;

	// Find E1000 Memory Base-Address
	// Base-Address in BAR0 on 32-bit systems and in BAR0/BAR1 on 64-bit systems
	// TODO: Support for 64-bit systems
	base_address_ = (volatile UInt8*) (cs->bar0 & 0xFFFFFFF0);

	regs_.ctrl   = (volatile UInt32*) (base_address_ + 0x0);
	regs_.ctrl_ext = (volatile UInt32*) (base_address_ + 0x18);
	regs_.status = (volatile UInt32*) (base_address_ + 0x8);
	regs_.eerd   = (volatile UInt32*) (base_address_ + 0x14);
	regs_.ledctl = (volatile UInt32*) (base_address_ + 0xe00);
	regs_.fextnvm = (volatile UInt32*) (base_address_ + 0x0028);

	regs_.ims    = (volatile UInt32*) (base_address_ + 0xD0);
	regs_.imc    = (volatile UInt32*) (base_address_ + 0xD8);
	regs_.icr    = (volatile UInt32*) (base_address_ + 0xC0);

	regs_.txcw   = (volatile UInt32*) (base_address_ + 0x178);
	regs_.tctl   = (volatile UInt32*) (base_address_ + 0x400);
	regs_.tipg   = (volatile UInt32*) (base_address_ + 0x410);
	regs_.tdba   = (volatile TransmitDescriptor**) (base_address_ + 0x3800);
	regs_.tdbal  = (volatile UInt32*) (base_address_ + 0x3800);
	regs_.tdbah  = (volatile UInt32*) (base_address_ + 0x3804);
	regs_.tdlen  = (volatile UInt32*) (base_address_ + 0x3808);

	regs_.tdh    = (volatile UInt32*) (base_address_ + 0x3810);
	regs_.tdt    = (volatile UInt32*) (base_address_ + 0x3818);

	regs_.txdctl = (volatile UInt32*) (base_address_ + 0x3828);



	regs_.rctl   = (volatile UInt32*) (base_address_ + 0x0100);
	regs_.rdba   = (volatile ReceiveDescriptor**) (base_address_ + 0x2800);
	regs_.rdbal  = (volatile UInt32*) (base_address_ + 0x2800);
	regs_.rdbah  = (volatile UInt32*) (base_address_ + 0x2804);
	regs_.rdlen  = (volatile UInt32*) (base_address_ + 0x2808);

	regs_.rdh    = (volatile UInt32*) (base_address_ + 0x2810);
	regs_.rdt    = (volatile UInt32*) (base_address_ + 0x2818);

	regs_.rxdctl = (volatile UInt32*) (base_address_ + 0x2828);

	regs_.rxcsum = (volatile UInt32*) (base_address_ + 0x5000);

	regs_.ra     = (volatile Address*) (base_address_ + 0x5400);
	regs_.mta     = (volatile UInt32*) (base_address_ + 0x5200);
}

void E1000::init() {

	printf("Initialisiere Register...\n");
	init_regs();

	// Start-Address of TX-descriptor-buffer (16-byte aligned)
	// Workaround for QEMU: it's impossible to read the TDBA[LH] registers!
	// TODO: __attribute__((align(16)) benutzen
	txd_buffer_start_ = (volatile TransmitDescriptor*) ((UInt64) (txd_buffer_+1) & 0xFFFFFFFFFFFFFFF0ULL);

	rxd_buffer_start_ = (volatile ReceiveDescriptor*) ((UInt64) (rxd_buffer_+1) & 0xFFFFFFFFFFFFFFF0ULL);

	//*regs_.status = 0;

	// disable power-down
	//*regs_.ctrl_ext &= ~(1 << 20);

	printf("Fuehre Reset durch...\n");
	*regs_.ctrl |= ( 1 << 26 ) | ( 1 << 31); // Reset + Link-Reset

	delay_1us();

	//*regs_.ctrl &= ~( 1 << 31 ); // Release Link-Reset

	// for qemu:
	// while (! (*regs_.status & (1 << 31)) ); // Warten, bis Reset fertig

	// for Optiplex 755:
	//while (*regs_.ctrl & ((1 << 26) | (1 << 31)) ); // Warten, bis Reset fertig

	//*regs_.ctrl &= ~(( 1 << 3 ) | ( 1 << 11 ) | ( 1 << 31 ));
	//*regs_.ctrl |= ( 1 << 5 ) | ( 1 << 6 );

	//printf("Warte auf Link...\n");
	//*regs_.ctrl = ( 1 << 0 ) | ( 1 << 5 ) | ( 1 << 6 ) | ( 10 << 8 ) ;


	// for qemu:
	// while (! (*regs_.status & ( 1 << 10)) ); // Warten, bis Link up

	// for Optiplex 755:
	//while (! (*regs_.status & ( 1 << 1)) ); // Warten, bis Link up

	// for Optiplex 755:
	//*regs_.ctrl = 	(1 << 11) | (1 << 12) | ( 1 << 0) | (1 << 8);

	// disable watchdog
	//*regs_.fextnvm = (1 << 10);

	//////////////
	// Receiver //
	//////////////
	// TODO: Intel 82566 (Dell Optiplex 755) hat keinen EEPROM...
	//setAddress( getDefaultAddress() );
	setAddress(0x112233445566ULL);

	for (int i = 0; i < 128; ++i) {
		regs_.mta[i] = 0;
	}

	*regs_.rdbal = 0;
	*regs_.rdbah = 0;

	*regs_.rdba = rxd_buffer_start_;
	*regs_.rdlen = RXD_BUFFERSIZE * sizeof(ReceiveDescriptor);

	*regs_.rdh = 0;
	*regs_.rdt = 0;

	*regs_.rxdctl |= ( 1 << 24 );

	// initialize descriptors for receive buffer
	for (unsigned i = 0; i < RXD_BUFFERSIZE; ++i) {
		rxd_buffer_start_[i].address = (UInt64) &rx_buffer_[i * 2048];
		rxd_buffer_start_[i].length  = 2048;

		rxd_buffer_start_[i].checksum = 0;
		rxd_buffer_start_[i].error = 0;
		rxd_buffer_start_[i].status = 0;
		rxd_buffer_start_[i].special = 0;
	}

	*regs_.rdt = RXD_BUFFERSIZE-1;

	//                EN           BAM      UPE (promisc)   SECRC (strip CRC field)
	*regs_.rctl= ( 1 << 1 ) | ( 1 << 15 ) | ( 1 << 3 )   |     ( 0 << 26 );
  //Qemu ignores SECRC=1 :-(


	/////////////////
	// Transmitter //
	/////////////////
	//*regs_.txcw = (1 << 31);

	*regs_.txdctl = (1 << 16) | (1 << 24);
	//*regs_.txdctl = ( 1 << 24 );

	*regs_.tdbal = 0;
	*regs_.tdbah = 0;

	// QEMU Workaround
	*regs_.tdba = txd_buffer_start_;

	*regs_.tdlen = TXD_BUFFERSIZE * sizeof(TransmitDescriptor);

	*regs_.tdh   = 0;
	*regs_.tdt   = 0;

	//             EN (enable TX)   PSP (padding)  CT (Collision Threshold)     COLD (collision distance)   RTLC (Retransmit on late collision
	*regs_.tctl  |= ( 1 << 1 )     |  ( 1 << 3 )   |   ( 0 << 4)                    | ( 0 << 12 )            | ( 1 << 24 );
	//*regs_.tctl  = ( 1 << 1 )     |  ( 1 << 3 )   |   ( 0x0F << 4)            | ( 0x3F << 12 )            | ( 1 << 24 );
	//*regs_.tctl  = ( 1 << 1 )     |  ( 1 << 3 )   |   ( 0x0F << 4)            | ( 0x3F << 12 )            | ( 1 << 24 );

	//*regs_.tctl |= (1 << 1) | (1 << 3);

	// TODO: Was für ein IPG?
	// Inter-Packet-Gap
	//*regs_.tipg  = ( 10 << 0 ) | ( 10 << 10 ) | ( 10 << 20 );
	//*regs_.tipg  = ( 10 << 0 ) | ( 10 << 10 ) | ( 10 << 20 );

	// LED Test - alle LED einschalten
	//*regs_.ledctl = (14 << 0) | (14 << 8) | (14 << 16);

	//*regs_.ctrl_ext = (1 << 28);

  printf("Initialisierung der Intel e1000 fertig!\n");

  // Register-Dump
  //printf("CTRL: 0x%x\n", *regs_.ctrl);
  //printf("STATUS: 0x%x\n", *regs_.status);
  //printf("TCTL: 0x%x\n", *regs_.tctl);
  //printf("TIDV: 0x%x\n", *regs_.tidv);

  //for (;;);
}

UInt16 E1000::readEEPROM(UInt8 addr) {
	*regs_.eerd = ( 1 << 0 ) | ( addr << 8 );

	while (! (*regs_.eerd & ( 1 << 4)) );

	return *regs_.eerd >> 16;
}

UInt64 E1000::getDefaultAddress() {
	UInt64 addr = 0;

	UInt16* addr16 = (UInt16*) &addr;

	// read EEPROM 00h-02h
	for (int i = 0; i < 3; ++i) {
		addr16[i] = readEEPROM(i);
	}

	return addr;
}


void E1000::enableInt() {
	//*regs_.ims = (1 << TXQE) | (1 << RXDMT0) | (1 << RXO) | (1 << RXT0) | (1 << RXDMT0);
	//*regs_.ims = (1 << RXT0);

	// TODO: Minimieren
	*regs_.ims = (1 << RXDMT0) | (1 << RXO) | (1 << RXT0);
}

void E1000::enableInt(InterruptSource source) {
	*regs_.ims |= ( 1 << source );
}

void E1000::disableInt(InterruptSource source) {
	*regs_.imc &= ~( 1 << source );
}


}
}
