#ifndef __HW_DEV_E1000_H__
#define __HW_DEV_E1000_H__

#include "util/types.h"

#include "ciaoConfig.h"

#include "hw/dev/IOPort.h"
#include "hw/dev/PCI.h"

namespace hw {
namespace dev {

class E1000 {

public:
	struct TransmitDescriptor {

		union {
			struct {
				// legacy transmit descriptor
				UInt64 address;

				UInt16 length;
				UInt8 cso;
				UInt8 cmd;

				UInt8 sta : 4;
				UInt8 rsv : 4;
				UInt8 css;
				UInt16 special;
			} legacy __attribute__((__packed__));

			struct {
				// TCP/IP context descriptor
				UInt8  ipcss;
				UInt8  ipcso;
				UInt16 ipcse;

				UInt8  tucss;
				UInt8  tucso;
				UInt16 tucse;

				UInt32 paylen : 20;
				UInt32 dtyp : 4;
				UInt32 tucmd : 8;

				UInt16 sta : 4;     // already in legacy transmit descriptor structure
				UInt16 rsv : 4;     // already in legacy transmit descriptor structure
				UInt16 hdrlen : 8;
				UInt16 mss;
			} context __attribute__((__packed__));

			struct {
				// TCP/IP data descriptor
				UInt64 address;     // already in legacy transmit descriptor structure

				UInt32 dtalen : 20;
				UInt32 dtyp : 4;    // already in TCP/IP context transmit descriptor structure
				UInt32 dcmd : 8;

				UInt16 sta : 4;     // already in legacy transmit descriptor structure
				UInt16 rsv : 4;     // already in legacy transmit descriptor structure
				UInt16 popts : 8;
				UInt16 special;     // already in legacy transmit descriptor structure
			} data __attribute__((__packed__));

		} __attribute__((__packed__));

	} __attribute__((__packed__));

	struct ReceiveDescriptor {

		UInt64 address;
		UInt16 length;
		UInt16 checksum;
		UInt8  status;
		UInt8  error;
		UInt16 special;

	} __attribute__((__packed__));

	struct Address {

		UInt64 address : 48;
		UInt64 rest : 16;

	} __attribute__((__packed__));

	enum InterruptSource {
		TXDW = 0,
		TXQE = 1,
		LSC  = 2,
		RXSEQ    = 3,
		RXDMT0   = 4,
		RXO      = 6,
		RXT0     = 7,
		MDAC     = 9,
		RXCFG    = 10,
		PHYINT   = 12,
		GPI0     = 13,
		GPI1     = 14,
		TXD_LOW  = 15,
		SRPD     = 16
	};

public:
	static const UInt16 DEVICE = cfDEV_E1000_DEVICE;
	static const UInt16 VENDOR = cfDEV_E1000_VENDOR;

private:
	volatile UInt8* base_address_;

	struct {
		volatile UInt32* ctrl;
		volatile UInt32* ctrl_ext;
		volatile UInt32* status;
		volatile UInt32* eerd;
		volatile UInt32* ledctl;
		volatile UInt32* fextnvm;

		volatile UInt32* ims;
		volatile UInt32* imc;
		volatile UInt32* icr;

		volatile UInt32* tctl;
		volatile UInt32* tipg;
		volatile UInt32* txcw;
		volatile TransmitDescriptor* volatile* tdba;    // QEMU can't read this register!
		volatile UInt32* tdbal;
		volatile UInt32* tdbah;
		volatile UInt32* tdlen;             // QEMU can't read this register!
		volatile UInt32* tdh;
		volatile UInt32* tdt;
		volatile UInt32* txdctl;

		volatile UInt32* rctl;
		volatile ReceiveDescriptor* volatile* rdba;
		volatile UInt32* rdbal;
		volatile UInt32* rdbah;
		volatile UInt32* rdlen;
		volatile UInt32* rdh;
		volatile UInt32* rdt;
		volatile UInt32* rxdctl;
		volatile UInt32* rxcsum;

		volatile Address* ra;
		volatile UInt32* mta;
	} regs_;

private:

	static E1000 inst_;

	UInt8 vector;

	UInt16 CIAO_NOINLINE readEEPROM(UInt8 addr); // INLINE doesn't work? :-(

	void delay_1us() {
		// TODO: mindestens 1us warten...
		for (volatile int i = 0; i < 1000000; ++i);
	}

public:
	static E1000& Inst() { return inst_; }
	void init();
	void init_regs();

	UInt64 getDefaultAddress();

	//UInt64 getAddress() { return regs_.ra->address; }

	void setAddress(UInt64 addr) {
		UInt32* ptr = (UInt32*) regs_.ra;

		ptr[0] = addr & 0xFFFFFFFF;
		//                                   Addr-Select    Addr valid
		ptr[1] = (( addr >> 32 ) & 0xFFFF) | ( 1 << 16 ) | ( 1 << 31 );
	}

	// TODO: eindeutige Namen auch bei mehreren Interfaces?
	const char* getName() { return "Intel 82540EM or compatible"; }

	bool hasTransmitterHardwareChecksumming() { return false; }
	bool hasReceiverHardwareChecksumming() { return false; }

  unsigned char getType() { return 1; } //1 := Ethernet
  unsigned getMTU() { return 1500; } // Payload MTU (without link-layer header)
  const unsigned char* getAddress() { return (const unsigned char*)regs_.ra; }

	UInt8 getVector() { return vector; }

	void enableInt(); // enable "default" int.-sources
	void enableInt(InterruptSource source);
	void disableInt(InterruptSource source);

	UInt32 getICR() volatile { return *regs_.icr; }
	void   clearICR() { *regs_.icr = 0xFFFFFFFF; }
};

}
}

#endif // __HW_DEV_E1000_H__
