#ifndef __HW_DEV_PCI_H__
#define __HW_DEV_PCI_H__

#include "util/types.h"

#include "hw/dev/IOPort.h"

#include "ciaoConfig.h"

namespace hw {
namespace dev {

struct PCIConfigSpace {

	UInt8 bus;
	UInt8 unit;
	UInt8 function;

	UInt16 device;
	UInt16 vendor;

	union {
		UInt32 classcode;

		struct {
			UInt8 progif;
			UInt8 subclasscode;
			UInt8 classcode8;
		};
	};

	UInt8  revision;
	UInt8  headertype;

	UInt32 bar0;
	UInt32 bar1;

	union {
		UInt32 bar2;

		struct {
			UInt8 primary_bus_number;
			UInt8 secondary_bus_number;
			UInt8 subordinate;
			UInt8 secondary_latency_timer;
		};
	};

	union {
		UInt32 bar3;

		struct {
			UInt8 io_base;
			UInt8 io_limit;
			UInt16 secondary_status;
		};
	};

	union {
		UInt32 bar4;

		struct {
			UInt16 memory_base;
			UInt16 memory_limit;
		};
	};

	union {
		UInt32 bar5;

		struct {
			UInt16 prefetch_memory_base;
			UInt16 prefetch_memory_limit;
		};
	};

	UInt32 rom_base;

	UInt8 interrupt_pin;
	UInt8 interrupt_line;
};

class PCI {
private:

	static PCI inst_;

	static const int CONFIGSPACE_MAXSIZE    = cfDEV_PCI_CONFIGSPACE_MAXSIZE;
	static const int IO_PORT_CONFIG_ADDRESS = cfDEV_PCI_CONFIG_ADDRESS;
	static const int IO_PORT_CONFIG_DATA    = cfDEV_PCI_CONFIG_DATA;

	PCIConfigSpace configspace_[CONFIGSPACE_MAXSIZE];

	int configspace_entries_;


	IOPort config_address_;
	IOPort config_data_;

private:
	// helper functions
	UInt32 getConfigSpaceData(unsigned bus, unsigned unit, unsigned function, unsigned offset) {
		config_address_.outl(0x80000000 | bus << 16 | unit << 11 | function << 8 | offset << 2);
		return config_data_.inl();
	}

public:

	static PCI& Inst() { return inst_; }

	static int getConfigSpaceMaxSize() { return CONFIGSPACE_MAXSIZE; }
	int getConfigSpaceEntries() { return configspace_entries_; }

	void init();

	void enumerate_bus(int secondary, int& next_free_busnum);

	const PCIConfigSpace* getConfigSpaceEntry(UInt16 device, UInt16 vendor) {
		for (int i = 0; i < getConfigSpaceEntries(); ++i) {
			if (configspace_[i].device == device && configspace_[i].vendor == vendor)
				return &configspace_[i];
		}

		return 0;
	}

	const PCIConfigSpace* getConfigSpaceEntryByClass(UInt8 classcode, UInt8 subclass) {
		for (int i = 0; i < getConfigSpaceEntries(); ++i) {
			if (configspace_[i].classcode8 == classcode && configspace_[i].subclasscode == subclass)
				return &configspace_[i];
		}

		return 0;
	}

	// for debugging
	const PCIConfigSpace& getConfigSpaceEntry(int index) { return configspace_[index]; }

};

}
}

#endif // __HW_DEV_PCI_H__
