/** \author Christian Meier
 *  \version $Id: ListManager.h 995 2009-08-11 14:26:31Z mike $
 */

#ifndef HW_DEV_CAN_MULTICAN_LIST_MANAGER_H
#define HW_DEV_CAN_MULTICAN_LIST_MANAGER_H

#include "os/com/ComStack_Types.h"
#include "hw/dev/can/HwObjectIdType.h"
#include "Cfg.h"
#include "RegTypes.h"
#include "Addr.h"

namespace hw { namespace dev { namespace can { namespace multican {

class ListManager {
public:
	void
	Init (const ConfigType * p_Config) {
		InitLists();
	}

	static ListManager & Inst() {
		return * (ListManager *) (HW_DEV_CAN_MULTICAN_MODULE_BASE_ADDRESS
								  + HW_DEV_CAN_MULTICAN_LIST_MANAGER_OFFSET);
	}

	enum ReturnType {
		LIST_MANAGER_OK,
		LIST_MANAGER_NOT_OK,
		LIST_MANAGER_TIMEOUT
	};

	enum InsertionType {
		INSERT_BEFORE,
		INSERT_AFTER
	};

	inline ReturnType WaitUntilReady(void) {
		for(uint32 i = 0; i < cfHW_DEV_CAN_MULTICAN_TIMEOUT_DURATION_FACTOR; ++i) {
			if(_PANCTR.bits.BUSY == 0) return LIST_MANAGER_OK;
		}
		return LIST_MANAGER_TIMEOUT;
	}

	ReturnType GetFirstHwObjHndl(uint8 ListHndl, HwObjectIdType * p_HwObjHndl) {
		if(_LIST[ListHndl].bits.EMPTY) {
			return LIST_MANAGER_NOT_OK;
		} else {
			*p_HwObjHndl = _LIST[ListHndl].bits.BEGIN;
			return LIST_MANAGER_OK;
		}
	}

	ReturnType GetListSize(uint8 ListHndl, uint8 * p_Size) {
		*p_Size = _LIST[ListHndl].bits.SIZE;
		return LIST_MANAGER_OK;
	}

	ReturnType GetLastHwObjHndl(uint8 ListHndl, HwObjectIdType * p_HwObjHndl) {
		if(_LIST[ListHndl].bits.EMPTY) {
			return LIST_MANAGER_NOT_OK;
		} else {
			*p_HwObjHndl = _LIST[ListHndl].bits.END;
			return LIST_MANAGER_OK;
		}
	}

	ReturnType InitLists(void) {
		// Wait to accept new command
		if(WaitUntilReady() == LIST_MANAGER_TIMEOUT) {
			return LIST_MANAGER_TIMEOUT;
		}

		// Issue Command
		IssueCmd(INIT_LISTS);

		// Wait until result is available
		if(WaitUntilResultReady() == LIST_MANAGER_TIMEOUT) {
			return LIST_MANAGER_TIMEOUT;
		}

		// Check if error occured
		if(ErrorOccured()) {
			return LIST_MANAGER_NOT_OK;
		} else {
			return LIST_MANAGER_OK;
		}
	}

	ReturnType StaticAllocate(uint8 TargetListHndl, HwObjectIdType HwObjHndl) {
		// Wait until parameters can be written
		if (WaitUntilResultReady() == LIST_MANAGER_TIMEOUT) {
			return LIST_MANAGER_TIMEOUT;
		};

		// Write parameters
		_PANCTR.bits.PANAR2 = TargetListHndl;
		_PANCTR.bits.PANAR1 = HwObjHndl;

		// Wait to accept new command
		if(WaitUntilReady() == LIST_MANAGER_TIMEOUT) {
			return LIST_MANAGER_TIMEOUT;
		}

		// Issue command
		IssueCmd(STATIC_ALLOCATE);

		return LIST_MANAGER_OK;
	}

	ReturnType DynamicAllocate(uint8 TargetListHndl, HwObjectIdType * p_SrcHwObjHndl) {
		// Wait until parameters can be written
		if (WaitUntilResultReady() == LIST_MANAGER_TIMEOUT) {
			return LIST_MANAGER_TIMEOUT;
		};

		// Write parameter
		_PANCTR.bits.PANAR2 = TargetListHndl;

		// Wait to accept new command
		if (WaitUntilReady() == LIST_MANAGER_TIMEOUT) {
			return LIST_MANAGER_TIMEOUT;
		}

		// Issue command
		IssueCmd(STATIC_ALLOCATE);

		// Wait until result is available
		if (WaitUntilResultReady() == LIST_MANAGER_TIMEOUT) {
			return LIST_MANAGER_TIMEOUT;
		}

		// Check if error occured
		if (ErrorOccured()) {
			return LIST_MANAGER_NOT_OK;
		}

		// Write output
		*p_SrcHwObjHndl = _PANCTR.bits.PANAR1;

		return LIST_MANAGER_OK;
	}

	ReturnType StaticInsert(InsertionType Pos, HwObjectIdType DestHwObjHndl, HwObjectIdType SrcHwObjHndl) {
		// Wait until parameters can be written
		if (WaitUntilResultReady() == LIST_MANAGER_TIMEOUT) {
			return LIST_MANAGER_TIMEOUT;
		};

		// Write parameters
		_PANCTR.bits.PANAR2 = DestHwObjHndl;
		_PANCTR.bits.PANAR1 = SrcHwObjHndl;

		// Wait to accept new command
		if (WaitUntilReady() == LIST_MANAGER_TIMEOUT) {
			return LIST_MANAGER_TIMEOUT;
		}

		// Issue command
		if (Pos == INSERT_BEFORE) {
			IssueCmd(STATIC_INSERT_BEFORE);
		} else {
			IssueCmd(STATIC_INSERT_BEHIND);
		}

		return LIST_MANAGER_OK;
	}

	// Seems to be that we don't need this for ciao AUTOSAR COM stack, but we for now let it at this position. --> Later configurable via pure::variants
	ReturnType DynamicInsert(InsertionType Pos, HwObjectIdType DestHwObjHndl, HwObjectIdType * p_SrcHwObjHndl) {
		// Wait until parameters can be written
		if (WaitUntilResultReady() == LIST_MANAGER_TIMEOUT) {
			return LIST_MANAGER_TIMEOUT;
		};

		// Write parameters
		_PANCTR.bits.PANAR2 = DestHwObjHndl;

		// Wait to accept new command
		if (WaitUntilReady() == LIST_MANAGER_TIMEOUT) {
			return LIST_MANAGER_TIMEOUT;
		}

		// Issue command
		if (Pos == INSERT_BEFORE) {
			IssueCmd(DYNAMIC_INSERT_BEFORE);
		} else {
			IssueCmd(DYNAMIC_INSERT_BEHIND);
		}

		// Wait until result is available
		if (WaitUntilResultReady() == LIST_MANAGER_TIMEOUT) {
			return LIST_MANAGER_TIMEOUT;
		}

		// Check if error occured
		if (ErrorOccured()) {
			return LIST_MANAGER_NOT_OK;
		}

		// Write output
		*p_SrcHwObjHndl = _PANCTR.bits.PANAR1;

		return LIST_MANAGER_OK;
	}

private:
	CAN_LISTm_t		_LIST[8];		//!< 0x0000 List Registers
	uint32			_dummy[41];		//!< 0x0020 Outside responsibility
	CAN_PANCTR_t	_PANCTR;		//!< 0x00C4 Panel Control Register

	enum Cmd {
		NO_OPERATION			= 0x00,
		INIT_LISTS				= 0x01,
		STATIC_ALLOCATE 		= 0x02,
		DYNAMIC_ALLOCATE		= 0x03,
		STATIC_INSERT_BEFORE	= 0x04,
		DYNAMIC_INSERT_BEFORE	= 0x05,
		STATIC_INSERT_BEHIND	= 0x06,
		DYNAMIC_INSERT_BEHIND	= 0x07
	};

	inline ReturnType WaitUntilResultReady(void) {
		for(uint32 i = 0; i < cfHW_DEV_CAN_MULTICAN_TIMEOUT_DURATION_FACTOR; ++i) {
			if(_PANCTR.bits.RBUSY == 0) return LIST_MANAGER_OK;
		}
		return LIST_MANAGER_TIMEOUT;
	}

	inline void IssueCmd (Cmd Command) {
		// Here we must wait until panel isn't anymore busy
		_PANCTR.bits.PANCMD = Command;
	}

	inline bool ErrorOccured(void) {
		return ((_PANCTR.bits.PANAR2 & 0x80) != 0);
	}
}; // class hw::dev::can::multican::ListManager

}}}}

#endif /* HW_DEV_CAN_MULTICAN_LIST_MANAGER_H */
