#ifndef HW_HAL_MPU_H_
#define HW_HAL_MPU_H_

#include <machine/intrinsics.h>
#include "stdinc.h"

#include <tc1796b/mpr.h>
#include <tc1796b/csfr.h>
#include <tc1796b/scu.h>

namespace hw { namespace hal {

class MPU
{
	typedef unsigned int tc_reg;

public:

	// typedef unsigned int addr;
	typedef Ptr addr;

	// addr operator(Ptr ptr) { return (addr) ptr; }

	enum DPerm {READ = 64, WRITE = 128};
	enum CPerm {EXECUTE = 128};
	typedef UInt8 Perms;

//Methods for all MPUs

	// to be adviced by the memory protection aspects
	static void init() {

		// init trusted (data)
		// trusted code uses the permissions specified in protection register set 0
		// we configure set 0 to enable full access to the entire address space
		MPR_DPMm_t_nonv dpm0;
		dpm0.reg=0;
		dpm0.bits.RE0=1;               // read enable range0
		dpm0.bits.WE0=1;               // write enable range0
		_mtcr( 0xe000 , dpm0.reg);     // write back mode register
		_mtcr( $dpr0_0l , 0x0);        // lower bound range0
		_mtcr( $dpr0_0u , 0xffffffff); // upper bound range0

		// init trusted (code)
		// we now use the second range here, since the first is likely to be
		// used as hardware breakpoint and for stepping. It *is* possible to
		// realize even single stepping solely in software, however, some
		// debuggers are rather unwilling to do so.
		MPR_CPMm_t_nonv cpm0;
		cpm0.reg=0;
		cpm0.bits.XE1=1;               // execution enable range1
		_mtcr( 0xe200 , cpm0.reg);     // write back mode register
		_mtcr( $cpr0_1l , 0x0);        // lower bound range1
		_mtcr( $cpr0_1u , 0xffffffff); // upper bound range1

		// init non trusted
		// make all non-trusted data-ranges readable
		/*TODO revisit, maybe it should be decided by a higher layer,
			which ranges are used at all and therefore readable*/
		MPR_DPMm_t_nonv dpm1;
		dpm1.reg=0;
		// (range0 will be set by initCommonRData)
		// dpm1.bits.RE1=1;    			//read enable range0
		// dpm1.bits.RE2=1;    			//read enable range0
		// dpm1.bits.RE3=1;    			//read enable range0
		_mtcr( 0xe080 , dpm1.reg);    	// write back mode register

		// set the big red main switch to ON
		SYSCON_t_nonv syscon;
		syscon.reg = _mfcr( $syscon );
		syscon.bits.PROTEN = 1;
		_mtcr( $syscon, syscon.reg );
		_isync();
	}

	// Configures the MPU to use the protection register set for non-trusted apps
	static void enable() {
		PSW_t_nonv psw;
		psw.reg = _mfcr( $psw );
		psw.bits.PRS = 1; // select Protection Register Set 1
		_mtcr( $psw, psw.reg );
	}

	// Configures the MPU to use the protection register set for trusted apps
	// does NOT actually disable the MP subsystem
	static void disable() {
		PSW_t_nonv psw;
		psw.reg = _mfcr( $psw );
		psw.bits.PRS = 0; // select Protection Register Set 0
		_mtcr( $psw, psw.reg );
	}

	// configures the currently active application data range, for which
	// data range 1 of register set 1 is used
	static void setPrivateDataRange(addr lower, addr upper) {
		_mtcr($dpr1_1l, (tc_reg) lower);
		_mtcr($dpr1_1u, (tc_reg) upper);
	}

	// configures the permissions for the application data range
	static void setPrivateDataPerms(Perms p) {
		MPR_DPMm_t_nonv dpm1;
		dpm1.reg = _mfcr(0xe080);  // get mode register
		UInt32 tmp = p;
		dpm1.reg |= (tmp<<8);
		_mtcr( 0xe080 , dpm1.reg); // write back mode register
	}


	// configures the currently active application code range, for which
	// code range 1 of register set 1 is used
	static void setPrivateCodeRange(addr lower, addr upper) {
		_mtcr($cpr1_1l, (tc_reg) lower);
		_mtcr($cpr1_1u, (tc_reg) upper);
	}

	// configures the permissions for the application code range
	static void setPrivateCodePerms(Perms p) {
		MPR_CPMm_t_nonv cpm1;
		cpm1.reg = _mfcr(0xe280);  // get mode register
		cpm1.reg |= ( ((UInt32) p) <<8);
		_mtcr( 0xe280 , cpm1.reg); // write back mode register
	}

	static void getStackRange(addr &lower, addr &upper) {
		lower = (addr) _mfcr($dpr1_2l);
		upper = (addr) _mfcr($dpr1_2u);
	}

	static addr getStackUpper() {
		return (addr) _mfcr($dpr1_2u);
	}

	// configures the stack range, for which data range 2 of
	// register set 1 is used
	static void setStackRange(addr lower, addr upper) {
		_mtcr($dpr1_2l, (tc_reg) lower);
		_mtcr($dpr1_2u, (tc_reg) upper);
	}

	// changes the upper bound of the stack range only
	static void setStackUpper(addr upper) {
		_mtcr($dpr1_2u, (tc_reg) upper);
	}

	// changes the lower bound of the stack range only
	static void setStackLower(addr lower) {
		_mtcr($dpr1_2l, (tc_reg) lower);
	}

	// sets the permissions of the stack range
	static void setStackPerms(Perms p) {
		MPR_DPMm_t_nonv dpm1;
		dpm1.reg = _mfcr(0xe080);  // get mode register
		dpm1.reg = (dpm1.reg & ~(0xFF<<16)) | (p<<16);
		_mtcr( 0xe080 , dpm1.reg); // write back mode register
	}

	// enables RW access to the stack range
	static void permitStack() {
		setStackPerms( READ | WRITE );
	}

	// configures permission for the stack range so that any access is forbidden
	static void forbidStack() {
		setStackPerms( 0 );
	}

	// configures the range for temorary message access, for which data range 3 of
	// register set 1 is used
	static void setMessageRange(addr lower, addr upper) {
		_mtcr($dpr1_3l, (tc_reg) lower);
		_mtcr($dpr1_3u, (tc_reg) upper);
	}

	static void getMessageRange(addr &lower, addr &upper) {
		lower = (addr) _mfcr($dpr1_3l);
		upper = (addr) _mfcr($dpr1_3u);
	}

	// configures permissions for the task-locals range
	static void setMessagePerms(Perms p) {
		MPR_DPMm_t_nonv dpm1;
		dpm1.reg = _mfcr(0xe080);   // get mode register
		dpm1.reg = (dpm1.reg & ~(0xFF<<24)) | (p<<24);
		_mtcr( 0xe080 , dpm1.reg);  // write back mode register
	}

	// configures permissions for the task-locals range
	static Perms getMessagePerms() {
		MPR_DPMm_t_nonv dpm1;
		dpm1.reg = _mfcr(0xe080);   // get mode register
		return (dpm1.reg >> 24);
	}

	// configures the common code range, i.e. a code region containing code
	// that is shared by all applications.
	// code range 1 of register set 1 is used for this purpose
	// we now use the second range here, since the first is likely to be
	// used as hardware breakpoint and for stepping. It *is* possible to
	// realize even single stepping solely in software, however, some
	// debuggers are rather unwilling to do so.
	static void initCommonCode(addr lower, addr upper) {
		MPR_CPMm_t_nonv cpm1;
		cpm1.reg = _mfcr(0xe280);        // get mode register

		cpm1.bits.XE1=1;                 // write enable range1
		_mtcr( 0xe280 , cpm1.reg);       // write back mode register
		_mtcr($cpr1_1l, (tc_reg) lower); // lower bound range1
		_mtcr($cpr1_1u, (tc_reg) upper); // upper bound range1
	}

	// configures the common code range to span the entire address space
	static void permitFullExecution() {
		initCommonCode(0, (Ptr) 0xFFFFFFFF);
	}

	// configures the common data range, i.e. data READABLE by all applications
	// range 0 of register set 1 is used for this purpose
	static void initCommonRData(addr lower, addr upper) {
		MPR_DPMm_t_nonv dpm1;
		dpm1.reg = _mfcr(0xe080);          // get mode register
		dpm1.bits.RE0=1;                   // write enable range0
		_mtcr( 0xe080 , dpm1.reg);         // write back mode register
		_mtcr( $dpr1_0l , (tc_reg) lower); // lower bound range0
		_mtcr( $dpr1_0u , (tc_reg) upper); // upper bound range0
	}

	// configures the common data range so that it spans the entire address space
	static void permitFullRead() {
		initCommonRData(0, (Ptr) 0xFFFFFFFF);
	}

	// leave trusted context to user mode 0 and enable the protection register
	// set 1 for non-trusted applications
	static void leaveTrusted() CIAO_INLINE {
		PSW_t_nonv psw;
		psw.reg = _mfcr( $psw );
		psw.bits.PRS = 1;
		psw.bits.IO = 0;  // user-0 mode
		_mtcr( $psw, psw.reg );
	}

	// leave trusted context to user mode 1 and enable the protection register
	// set 1 for non-trusted applications
	// peripheral access is still allowed for non-trusted apps in this mode
	static void leaveTrusted_drv() CIAO_INLINE {
		PSW_t_nonv psw;
		psw.reg = _mfcr( $psw );
		psw.bits.PRS = 1;
		psw.bits.IO = 1;  // user-1 mode
		_mtcr( $psw, psw.reg );
		_isync();
	}

	static bool isTrusted() CIAO_INLINE {
		PSW_t_nonv psw;
		psw.reg = _mfcr( $psw );
		return psw.bits.IO == 2;
	}


	////////////////////////////////////////////////////////////////////////////
	// FIXME this looks like enterTrusted and enterPrivileged need be swapped //
	////////////////////////////////////////////////////////////////////////////

	// change processor to privileged mode
	inline static void enterTrusted() CIAO_INLINE {
		_syscall(0);
	}

	// change processor to privileged mode and enable the MP context for trusted
	// applications (i.e. protection register set 0)
	inline static void enterPrivileged() CIAO_INLINE {
		_syscall(0);
		disable();
	}

	inline static void leavePrivileged_drv() CIAO_INLINE {
		PSW_t_nonv psw;
		psw.reg = _mfcr( $psw );
		psw.bits.IO = 1; // user-1 mode
		_mtcr( $psw, psw.reg );
		_isync();
	}

	inline static void leavePrivileged() CIAO_INLINE {
		PSW_t_nonv psw;
		psw.reg = _mfcr( $psw );
		psw.bits.IO = 0; // user-0 mode
		_mtcr( $psw, psw.reg );
		_isync();
	}
};

}} // hw::hal

#endif /*MPU_H_*/
