#ifndef STMTIMER0_H_
#define STMTIMER0_H_


#include "STMTimerBase.h"
#include "CPU.h"

namespace hw { namespace hal {

	using namespace hw::dev;

	/**
	 * @brief typically used to implement hal::SystemTimer
	 *
	 * @tc STM timers are built on top of the system timer module (STM). STMTimer0 is hard-wired
	 * to use the CMP0 compare-and-match register and to use hw::irq::STM_SRC0 as CPU interrupt SRN.
	 *
	 */
	class STMTimer0 : public STMTimerBase< hw::dev::STM::IRQ_SRC0 > {
		static STMTimer0 theSTMTimer0_;

		//! Performs start-time initializations. Only called by advice.
		void init();

		protected:
		Unit	period_;
		Unit	overflowSnapshot_;

		STMTimer0(){}

		public:
		typedef STMTimerBase< hw::dev::STM::IRQ_SRC0 > Super;

		//! Returns the one-and-only Timer0 instance
		static STMTimer0& Inst() {
			return theSTMTimer0_;
		}

		/**
		 * @brief  Windup timer as periodical timer.
		 *
		 * A periodical timer triggers an irq every \a timer units.
		 *
		 */
		CIAO_STD void windupPeriodical( Unit period );

		/**
		 * @brief  Windup timer as single shot
		 *
		 * A single shot timer triggers an ... TODO
		 */
		CIAO_STD void windupAndStartSingle( Unit period, Unit decisionBase);


		CIAO_STD void start();

		CIAO_STD void stop() {
			IRQ::disable();
			hw::dev::STM::ICR().bits.CMP0EN = 0;
		}

		CIAO_STD static Bool running() {
			return hw::dev::STM::ICR().reg & STM_ICR_CMP0EN_MASK;
		}

		CIAO_STD static void clear() {
			hw::dev::STM::ISRR().bits.CMP0IRR = 1;
			_dsync();     // ensure bit has been cleared (or do we need _isync() here?)
			IRQ::clear(); // also clear (possibly already pending) SRN flag
		}

		//! Returns true, if service request is pending.
		static Bool pending() {
			// test for pending internal IRQ should be enough,
			// as SRN flag should always be in line with it.
			return hw::dev::STM::ICR().reg & STM_ICR_CMP0IR_MASK;
		}

		//! Explicitly set service request pending flag.
		static void set() {
			// set internal IRQ flag, which automatically sets the SRN flag, if enabled
			hw::dev::STM::ISRR().bits.CMP0IRS = 1;
		}

		/**
		 * @brief Acknowledge timer event. To be called by the IRQ handler (prologue!).
		 *
		 * @todo Maybe we should provide an aspect for this.
		 */
		CIAO_STD void CIAO_INLINE irqAck() {
			if( period_ != 0 )
				hw::dev::STM::CMP0() += period_;
			else
				stop();

			// internal pending bit has to be cleared explicitly,
			// SRN pending bit has already been cleared by hardware
			hw::dev::STM::ISRR().bits.CMP0IRR = 1;
		}

		/**
		 * Provides support for checking for overflow of TIM<0> since last call to this
		 * method. If an overflow occured, the next call to this method will only return
		 * true if the next overflow has been detected.
		 */
		CIAO_STD Bool CIAO_INLINE checkOverflowAndClear () {
			Unit newSnapshot =  hw::dev::STM::TIM<6> ();
			if (newSnapshot != overflowSnapshot_) {
				overflowSnapshot_ = newSnapshot;
				return true;
			} else {
				return false;
			}
		}
	};
}} //namespaces
#endif /*STMTIMER0_H_*/
