/** \author Christian Meier
 *  \version $Id: Controller.h 995 2009-08-11 14:26:31Z mike $
 *
 *  One Controller coordinates the current state and transitions of the underlying
 *  hardware controller.
 */

#ifndef HW_HAL_CAN_CONTROLLER_H
#define HW_HAL_CAN_CONTROLLER_H

#include "os/com/ComStack_Types.h"
#include "hw/hal/can/ControllerConfigType.h"

#include "ControllerIdType.h"
#include "Cfg.h"

namespace hw { namespace hal { namespace can {

class Controller : private ControllerConfigType {
public:
	inline void
	Init(ControllerConfigType const * ControllerConfig) {
		*(ControllerConfigType *)this = *ControllerConfig;
		CtrlMode_ = CANIF_CS_STOPPED;
	}

	inline ControllerIdType
	Id() const {
		return this - Controllers_;
	}

	inline void
	InitController(uint8 ConfigurationIndex) {
		InitController_Delegator(&CtrlConfigs_[ConfigurationIndex]);
		CtrlMode_ = CANIF_CS_STOPPED;
	}

	inline StatusType
	SetControllerMode(ControllerModeType ControllerMode) {
		// What happens when current mode is uninit or undefined
		switch (CtrlMode_) {
		case CANIF_CS_UNINIT:	return E_NOT_OK;
		case CANIF_CS_STOPPED:	break;
		case CANIF_CS_STARTED:	break;
		case CANIF_CS_SLEEP:	break;
		default:				return E_NOT_OK;
		}

		// What happens for requested ControllerMode
		switch (ControllerMode) {
		case CANIF_CS_UNINIT:	return E_NOT_OK;
		case CANIF_CS_STOPPED:	return SetControllerMode_STOP();
		case CANIF_CS_STARTED:	return SetControllerMode_START();
		case CANIF_CS_SLEEP:	return SetControllerMode_SLEEP();
		default:				return E_NOT_OK;
		}
	}

	inline StatusType
	GetControllerMode(ControllerModeType * ControllerMode) {
		* ControllerMode = CtrlMode_;
		return E_OK;
	}

private:
	inline Controller() {}

	// Ugly workaround, currently I don't get a non-inline introduction of
	// SetControllerMode_Delegator working so we do it manually
	hw::dev::can::ReturnType
	SetControllerMode_Delegator_NonInline(hw::dev::can::StateTransitionType Transition);

	StatusType
	SetControllerMode_STOP() {
		if (SetControllerMode_Delegator_NonInline(hw::dev::can::CAN_T_STOP) != hw::dev::can::CAN_OK) {
			return E_NOT_OK;
		} else {
			CtrlMode_ = CANIF_CS_STOPPED;
			return E_OK;
		}
	}

	StatusType
	SetControllerMode_START() {
		if (CtrlMode_ == CANIF_CS_SLEEP) {
			if (SetControllerMode_Delegator_NonInline(hw::dev::can::CAN_T_STOP) != hw::dev::can::CAN_OK) {
				return E_NOT_OK;
			}
			CtrlMode_ = CANIF_CS_STOPPED;
		}

		if (SetControllerMode_Delegator_NonInline(hw::dev::can::CAN_T_START) != hw::dev::can::CAN_OK) {
			return E_NOT_OK;
		} else {
			CtrlMode_ = CANIF_CS_STARTED;
			return E_OK;
		}
	}

	StatusType
	SetControllerMode_SLEEP() {
		if (CtrlMode_ == CANIF_CS_STARTED) {
			if (SetControllerMode_Delegator_NonInline(hw::dev::can::CAN_T_STOP) != hw::dev::can::CAN_OK) {
				return E_NOT_OK;
			}
			CtrlMode_ = CANIF_CS_STOPPED;
		}

		if (SetControllerMode_Delegator_NonInline(hw::dev::can::CAN_T_SLEEP) != hw::dev::can::CAN_OK) {
			return E_NOT_OK;
		} else {
			return E_OK;
			CtrlMode_ = CANIF_CS_SLEEP;
		}
	}

	ControllerModeType CtrlMode_;

public:
	static void
	InitAll(ControllerConfigType const * ConfigArray, ControllerIdType ControllerMultiplicity) {
		ControllerIdType i = 0;

		// We assume ControllerMultiplicity <= cfHW_HAL_CAN_CONTROLLER_MULTIPLICITY
		for(; i < ControllerMultiplicity; ++i) {
			Controllers_[i].Init(&ConfigArray[i]);
		}

		ControllerMultiplicity_ = i;

		for(; i < cfHW_HAL_CAN_CONTROLLER_MULTIPLICITY; ++i) {
			Controllers_[i].SetControllerMode(CANIF_CS_STOPPED);
			Controllers_[i].CtrlMode_ = CANIF_CS_UNINIT;
		}
	}

	static Controller&
	Inst(ControllerIdType ControllerId) {
		return Controllers_[ControllerId];
	}

	static ControllerIdType
	Multiplicity() {
		return ControllerMultiplicity_;
	}

private:
	static Controller Controllers_[cfHW_HAL_CAN_CONTROLLER_MULTIPLICITY];
	static ControllerIdStorageType ControllerMultiplicity_;
};

}}}

#endif
