#ifndef BITMANIP_H
#define BITMANIP_H

#include "util/metaprog.h"
#include "util/types.h"

#include <stdio.h>

namespace bitmanip {

// Below the general implementations of the provided algorithms are provided
// in case of there is no special implementation of the respective functionality.

namespace impl {

/** ***************************************
 *  setBitAlgo
 */
template <typename INT_TYPE, bool BISECT, bool ARCH=true>
struct setBitAlgo {
	static inline INT_TYPE
	Do (INT_TYPE Value, UInt BitIndex) {
		return Value | (INT_TYPE)(((INT_TYPE)1) << BitIndex);
	}
};

template <typename INT_TYPE, bool ARCH>
struct setBitAlgo<INT_TYPE,true,ARCH> {
	typedef typename int_t<sizeof(INT_TYPE)/2,false>::T BisectValueType;
	typedef setBitAlgo<BisectValueType,(sizeof(BisectValueType) > sizeof(UInt)),ARCH> SmallerAlgo;

public:
	static inline INT_TYPE
	Do (INT_TYPE Value, UInt BitIndex) {
		BisectValueType Upper;
		BisectValueType Lower;

		if(BitIndex >= 4*sizeof(INT_TYPE)) {
			Upper = SmallerAlgo::Do (Upper, BitIndex - 4*sizeof(INT_TYPE));
			Lower = lowerHalf(Value);
		} else {
			Lower = SmallerAlgo::Do (Lower, BitIndex);
			Upper = upperHalf(Value);
		}
		return (((INT_TYPE)Upper)<<(4*sizeof(INT_TYPE))) | (INT_TYPE)Lower;
	}
};

/** ***************************************
 *  clearBitAlgo
 */
template <typename INT_TYPE, bool BISECT, bool ARCH=true>
struct clearBitAlgo {
	static inline INT_TYPE
	Do (INT_TYPE Value, UInt BitIndex) {
		return Value & ~(INT_TYPE)(((INT_TYPE)1) << BitIndex);
	}
};

template <typename INT_TYPE, bool ARCH>
struct clearBitAlgo<INT_TYPE,true,ARCH> {
	typedef typename int_t<sizeof(INT_TYPE)/2,false>::T BisectValueType;
	typedef clearBitAlgo<BisectValueType,(sizeof(BisectValueType) > sizeof(UInt)),ARCH> SmallerAlgo;

public:
	static inline INT_TYPE
	Do (INT_TYPE Value, UInt BitIndex) {
		BisectValueType Upper;
		BisectValueType Lower;

		if(BitIndex >= 4*sizeof(INT_TYPE)) {
			Upper = SmallerAlgo::Do (Upper, BitIndex - 4*sizeof(INT_TYPE));
			Lower = lowerHalf(Value);
		} else {
			Lower = SmallerAlgo::Do (Lower, BitIndex);
			Upper = upperHalf(Value);
		}
		return (((INT_TYPE)Upper)<<(4*sizeof(INT_TYPE))) | (INT_TYPE)Lower;
	}
};

/** ***************************************
 *  testBitAlgo
 */
template <typename INT_TYPE, bool BISECT, bool ARCH=true>
struct testBitAlgo {
	static inline Bool
	Do (INT_TYPE Value, UInt BitIndex) {
		return (Value & (INT_TYPE)(((INT_TYPE)1) << BitIndex)) != 0;
	}
};

template <typename INT_TYPE, bool ARCH>
struct testBitAlgo<INT_TYPE,true,ARCH> {
	typedef typename int_t<sizeof(INT_TYPE)/2,false>::T BisectValueType;
	typedef testBitAlgo<BisectValueType,(sizeof(BisectValueType) > sizeof(UInt)),ARCH> SmallerAlgo;

public:
	static inline INT_TYPE
	Do (INT_TYPE Value, UInt BitIndex) {
		BisectValueType Upper;
		BisectValueType Lower;

		if(BitIndex >= 4*sizeof(INT_TYPE)) {
			Upper = SmallerAlgo::Do (Upper, BitIndex - 4*sizeof(INT_TYPE));
			Lower = lowerHalf(Value);
		} else {
			Lower = SmallerAlgo::Do (Lower, BitIndex);
			Upper = upperHalf(Value);
		}
		return (((INT_TYPE)Upper)<<(4*sizeof(INT_TYPE))) | (INT_TYPE)Lower;
	}
};

/** ***************************************
 *  highest1BitAlgo
 */
template <typename INT_TYPE, bool BISECT, bool ARCH=true>
struct highest1BitAlgo {
	static const UInt BIT_LENGTH = sizeof(INT_TYPE) * __CHAR_BIT__;

public:
	static UInt
	Do (INT_TYPE Value) {
		/*	O(log n) algorithm */
		UInt LowerBitIndex = 0;
		UInt UpperBitIndex = BIT_LENGTH - 1;
		UInt MiddleLowerBitIndex;
		INT_TYPE Comp;

		while (LowerBitIndex != UpperBitIndex) {
			MiddleLowerBitIndex = ((UpperBitIndex - LowerBitIndex) >> 1) + LowerBitIndex;
			Comp = 2 << MiddleLowerBitIndex;

			if (Value >= Comp) {
				LowerBitIndex = MiddleLowerBitIndex + 1;
			} else {
				UpperBitIndex = MiddleLowerBitIndex;
			}
		}
		return LowerBitIndex;

		/*	Maybe should be configurable:
			slow O(n) but smaller algorithm
		UInt BitIndex = BIT_LENGTH - 1;

		for(; BitIndex; --BitIndex) {
			if (Value & (((INT_TYPE)1) << BitIndex)) break;
		}
		// As result for Value==0 is unspecified we can savely return BitIndex here.
		return BitIndex;
		*/
	}
};

template <typename INT_TYPE, bool ARCH>
struct highest1BitAlgo  <INT_TYPE, true, ARCH> {
	typedef typename int_t<sizeof(INT_TYPE)/2,false>::T BisectValueType;
	typedef highest1BitAlgo<BisectValueType, (sizeof(BisectValueType) > sizeof(UInt))> SmallerAlgo;

public:
	static inline UInt
	Do (INT_TYPE Value) {
		BisectValueType Upper = upperHalf(Value);

		if (Upper) {
			return SmallerAlgo::Do(Upper) + 4*sizeof(INT_TYPE);
		} else {
			return SmallerAlgo::Do(lowerHalf(Value));
		}
	}
};

/** ***************************************
 *  highest0BitAlgo
 */
template <typename INT_TYPE, bool BISECT, bool ARCH=true>
class highest0BitAlgo {
	typedef highest1BitAlgo<INT_TYPE, false> ProxyAlgo;

public:
	static inline UInt
	Do (INT_TYPE Value) {
		return ProxyAlgo::Do (~Value);
	}
};

template <typename INT_TYPE, bool ARCH>
struct highest0BitAlgo <INT_TYPE, true, ARCH> {
	typedef typename int_t<sizeof(INT_TYPE)/2,false>::T BisectValueType;
	typedef highest0BitAlgo<BisectValueType, (sizeof(BisectValueType) > sizeof(UInt))> SmallerAlgo;

public:
	static inline UInt
	Do (INT_TYPE Value) {
		BisectValueType Upper = upperHalf(Value);

		if ((BisectValueType)~Upper) {
			return SmallerAlgo::Do(Upper) + 4*sizeof(INT_TYPE);
		} else {
			return SmallerAlgo::Do(lowerHalf(Value));
		}
	}
};

/** ***************************************
 *  lowest1BitAlgo
 */
template <typename INT_TYPE, bool BISECT, bool ARCH=true>
struct lowest1BitAlgo {
	static const UInt BIT_LENGTH = sizeof(INT_TYPE) * __CHAR_BIT__;

public:
	static UInt
	Do (INT_TYPE Value) {
		/*	O(log n) algorithm */
		UInt LowerBitIndex = 0;
		UInt UpperBitIndex = BIT_LENGTH - 1;
		UInt MiddleLowerBitIndex;
		INT_TYPE Mask;

		while (LowerBitIndex != UpperBitIndex) {
			MiddleLowerBitIndex = ((UpperBitIndex - LowerBitIndex) >> 1) + LowerBitIndex;
			Mask = (2 << MiddleLowerBitIndex) - 1;

			if (Value & Mask) {
				UpperBitIndex = MiddleLowerBitIndex;
			} else {
				LowerBitIndex = MiddleLowerBitIndex + 1;
			}
		}
		return LowerBitIndex;

		/*	Maybe should be configurable:
			slow O(n) but smaller algorithm
		UInt BitIndex = 0;
		for(; BitIndex < BIT_LENGTH - 1; ++BitIndex) {
			if (Value & (((INT_TYPE)1) << BitIndex)) break;
		}
		return BitIndex;
		*/
	}
};

template <typename INT_TYPE, bool ARCH>
struct lowest1BitAlgo<INT_TYPE, true, ARCH> {
	typedef typename int_t<sizeof(INT_TYPE)/2,false>::T BisectValueType;
	typedef lowest1BitAlgo<BisectValueType, (sizeof(BisectValueType) > sizeof(UInt))> SmallerAlgo;

public:
	static inline UInt
	Do (INT_TYPE Value) {
		BisectValueType Lower = lowerHalf(Value);

		if (Lower) {
			return SmallerAlgo::Do(Lower) + 4*sizeof(INT_TYPE);
		} else {
			return SmallerAlgo::Do(upperHalf(Value));
		}
	}
};

/** ***************************************
 *  lowest0BitAlgo
 */
template <typename INT_TYPE, bool BISECT, bool ARCH=true>
class lowest0BitAlgo {
	typedef lowest1BitAlgo<INT_TYPE, false> ProxyAlgo;

public:
	static inline UInt
	Do (INT_TYPE Value) {
		return ProxyAlgo::Do (~Value);
	}
};

template <typename INT_TYPE, bool ARCH>
struct lowest0BitAlgo<INT_TYPE, true, ARCH> {
	typedef typename int_t<sizeof(INT_TYPE)/2,false>::T BisectValueType;
	typedef lowest0BitAlgo<BisectValueType, (sizeof(BisectValueType) > sizeof(UInt))> SmallerAlgo;

public:
	static inline UInt
	Do (INT_TYPE Value) {
		BisectValueType Lower = lowerHalf(Value);

		if ((BisectValueType)~Lower) {
			return SmallerAlgo::Do(Lower) + 4*sizeof(INT_TYPE);
		} else {
			return SmallerAlgo::Do(upperHalf(Value));
		}
	}
};

/** ***************************************
 *  twosComplementAlgo
 */
template <class INT_TYPE, bool ARCH=true>
struct twosComplementAlgo {
	static inline INT_TYPE
	Do (INT_TYPE Value) {
		// As this should be a plattform independent implementation
		// we don't know whether we work on a one's or two's complement
		// architecture. Because using just (~Value) + 1 on one's comp. arch.
		// for Value=0 would lead to (~0)+1 to b11..11+1 to b00..01 which
		// is not two's comp. of 0 we must catch the case of Value=0
		if (Value == (INT_TYPE)0) return 0;
		return ((INT_TYPE)~Value) + (INT_TYPE)1;
	}
};

/** ***************************************
 *  extractAlgo
 */
template<class INT_TYPE, bool ARCH=true>
struct extractAlgo {
	static INT_TYPE
	Do (INT_TYPE Value, UInt BitIndex, UInt BitLength) {
		INT_TYPE Mask = ~0;
		if(BitLength < __CHAR_BIT__*sizeof(INT_TYPE)) {
			Mask = Mask << BitLength;
			Mask = ~Mask;
		}
		return ((INT_TYPE)(Value >> BitIndex)) & Mask;
	}
};

/** ***************************************
 *  insertAlgo
 */
template<class INT_TYPE, bool ARCH=true>
struct insertAlgo {
	static INT_TYPE
	Do (INT_TYPE Value, INT_TYPE NewValue, UInt BitIndex, UInt BitLength) {
		INT_TYPE Mask = ~0;
		if(BitLength < __CHAR_BIT__*sizeof(INT_TYPE)) {
			Mask = Mask << BitLength;
			Mask = ~Mask;
			Mask = Mask << BitIndex;
		}
		return (Value & ~Mask) | (((INT_TYPE)(NewValue<<BitIndex)) & Mask);
	}
};

/** ***************************************
 *  insertBitAlgo
 */
template<class INT_TYPE, bool ARCH=true>
struct insertBitAlgo {
	static inline INT_TYPE
	Do (INT_TYPE Value, INT_TYPE NewValue, UInt BitIndex) {
		INT_TYPE Mask = ((INT_TYPE)1) << BitIndex;
		return (Value & ~Mask) | (((INT_TYPE)(NewValue<<BitIndex)) & Mask);
	}
};

}} // namespace bitmanip::impl

// Architecture dependent specializations
#include "util/bitmanip_arch.h"

// This is the "public interface", only this interface should be called from outside
// This interface maps all calls with signed integers to calls with unsigned integers.

namespace bitmanip {

// Normally for debugging only:
//#define BITMANIP_INLINE __attribute__((noinline))
#define BITMANIP_INLINE inline

// returns the index of the most significant bit that is 1
template< class INT_TYPE >
BITMANIP_INLINE UInt
highest1Bit (INT_TYPE Value) {
	return impl::highest1BitAlgo<typename int_t_to_unsigned<INT_TYPE>::T,(sizeof(INT_TYPE) > sizeof(UInt))>::Do((typename int_t_to_unsigned<INT_TYPE>::T)Value);
}

template< class INT_TYPE >
BITMANIP_INLINE UInt
lowest1Bit (INT_TYPE Value) {
	return impl::lowest1BitAlgo<typename int_t_to_unsigned<INT_TYPE>::T,(sizeof(INT_TYPE) > sizeof(UInt))>::Do((typename int_t_to_unsigned<INT_TYPE>::T)Value);
}

template< class INT_TYPE >
BITMANIP_INLINE UInt
highest0Bit (INT_TYPE Value) {
	return impl::highest0BitAlgo<typename int_t_to_unsigned<INT_TYPE>::T,(sizeof(INT_TYPE) > sizeof(UInt))>::Do((typename int_t_to_unsigned<INT_TYPE>::T)Value);
}

template< class INT_TYPE >
BITMANIP_INLINE UInt
lowest0Bit (INT_TYPE Value) {
	return impl::lowest0BitAlgo<typename int_t_to_unsigned<INT_TYPE>::T,(sizeof(INT_TYPE) > sizeof(UInt))>::Do((typename int_t_to_unsigned<INT_TYPE>::T)Value);
}

template<class INT_TYPE>
BITMANIP_INLINE INT_TYPE
clearBit (INT_TYPE Value, UInt BitIndex) {
	return impl::clearBitAlgo<typename int_t_to_unsigned<INT_TYPE>::T,(sizeof(INT_TYPE) > sizeof(UInt))>::Do((typename int_t_to_unsigned<INT_TYPE>::T)Value, BitIndex);
}

template<class INT_TYPE>
BITMANIP_INLINE INT_TYPE
setBit (INT_TYPE Value, UInt BitIndex) {
	return impl::setBitAlgo<typename int_t_to_unsigned<INT_TYPE>::T,(sizeof(INT_TYPE) > sizeof(UInt))>::Do((typename int_t_to_unsigned<INT_TYPE>::T)Value, BitIndex);
}

template<class INT_TYPE>
BITMANIP_INLINE Bool
testBit (INT_TYPE Value, UInt BitIndex) {
	return impl::testBitAlgo<typename int_t_to_unsigned<INT_TYPE>::T,(sizeof(INT_TYPE) > sizeof(UInt))>::Do((typename int_t_to_unsigned<INT_TYPE>::T) Value, BitIndex);
}

template<class INT_TYPE>
BITMANIP_INLINE Bool
testAndClearBit (INT_TYPE * ValuePtr, UInt BitIndex) {
	INT_TYPE Tmp = *ValuePtr & (1<<BitIndex);
	*ValuePtr &= ~Tmp;
	return Tmp;
}

template<class INT_TYPE>
BITMANIP_INLINE INT_TYPE
twosComplement(INT_TYPE Value) {
	return impl::twosComplementAlgo<typename int_t_to_unsigned<INT_TYPE>::T>::Do((typename int_t_to_unsigned<INT_TYPE>::T)Value);
}

template<class INT_TYPE>
BITMANIP_INLINE INT_TYPE
insert (INT_TYPE Value, INT_TYPE NewValue, UInt BitIndex, UInt BitLength) {
	return impl::insertAlgo<INT_TYPE>::Do(Value, NewValue, BitIndex, BitLength);
}

template<class INT_TYPE>
BITMANIP_INLINE INT_TYPE
insertBit (INT_TYPE Value, INT_TYPE NewValue, UInt BitIndex) {
	return impl::insertBitAlgo<INT_TYPE>::Do(Value, NewValue, BitIndex);
}

// Extracts some bits from a value starting with bit BitIndex and upwards BitLength bits
template<class INT_TYPE>
BITMANIP_INLINE INT_TYPE
extract (INT_TYPE Value, UInt BitIndex, UInt BitLength) {
	// !!! no conversion to unsigned types here (sign extension)
	return impl::extractAlgo<INT_TYPE>::Do(Value, BitIndex, BitLength);
}

// Extracts an number of bits compile time from the lower end of an integer
template<typename INT_TYPE>
BITMANIP_INLINE INT_TYPE
extract_low(INT_TYPE Value, UInt NumBits) {
	Value <<= sizeof(INT_TYPE)*__CHAR_BIT__ - NumBits;
	// Depending on the type this will automatically make the sign extension
	Value >>= sizeof(INT_TYPE)*__CHAR_BIT__ - NumBits;
	return Value;
}

// Extracts an number of bits known at compile time from the lower end of an integer
template<typename INT_TYPE, int NUM_BITS>
BITMANIP_INLINE INT_TYPE
static_extract_low(INT_TYPE Value) {
	// This struct can extract signed, as well as unsigned data!
	struct {
		INT_TYPE Res : NUM_BITS;
	} Converter;
	Converter.Res = Value;
	return Converter.Res;
}

} // namespace bitmanip

#endif

