#ifndef __StaticAllocator_h__
#define __StaticAllocator_h__

#include "mem/BaseAllocator.h"
#include "util/container/Chain.h"
#include "util/types.h"

//! StaticAllocator  
/*! 
 *  Allows different container instances ( using the same type)
 *  to use the same area in memory for allocation.
 */
template<class T, int S, int I=0>
class StaticAllocator {

// public typedefs and enums
public:

	//! Type definition for instatiated type
	typedef T value_t;
	
	//! Type definition for pointer to intantiated type
	typedef Ptr pointer_t;
	
	typedef typename index_t_value<S>::T index_t;
	
	//! Type definition for this type of allocator
	typedef StaticAllocator<T,S,I> alloc_t;
	
	enum{SIZE=S};
	enum{ID=I};


// protected variables
protected:
	//! memory area for storing objects of \ref value_t
	static UInt8 _pool[ S * sizeof(value_t) ];

	//! array of pointers to objects of \ref value_t
	static value_t* _ptr[S];
	
	//! position in \ref _ptr pointing out the next free block of memory
	static typename index_t_value<S>::T  _pos;
	//static UInt16 _pos;
	
	//! indicates if Allocator is initialized
	static bool _initialized;
	
	
	
// public methods
public:
	//! constructor
	StaticAllocator()
	{
		if(!_initialized){
			for(int i=0;i<SIZE;++i){
				_ptr[i]=(value_t*) &(_pool[i*(sizeof(value_t))]);
			}
			_initialized=true;
		}
	}
	
	StaticAllocator(const StaticAllocator& alloc){}
	
	//! allocate memory area 
	/**
	 * \return Pointer of type  to memory location
	 */
 	pointer_t allocate();
	
	
	//! allocate memory area 
	/**
	 * \return Pointer of type  to memory location
	 */
 	pointer_t allocate(index_t size)//;
 	{
		return allocate();
	}
	
 	
 	//! free memory 
 	/**
 	 * \param obj is the adress of the object which should be freed
 	 */
	void free(pointer_t obj)//;
	{
			--_pos;
			_ptr[_pos]=static_cast<value_t* >(obj);
	}
	
	//! get number of used memory locations
	int used(){
		return _pos;
	}
};

template<class T, int S,int I>	
UInt8 StaticAllocator<T,S,I>::_pool[ S * sizeof(T) ];

template<class T, int S, int I>	
T* StaticAllocator<T,S,I>::_ptr[S];

/* _pos and _initialized should be initialized to zero. This works not on H8. So 
 * we are assuming that bss segment (where _pos is located) is initialized with 
 * zero */
template<class T, int S, int I>	
typename  index_t_value<S>::T StaticAllocator<T,S,I>::_pos;

template<class T, int S, int I>	
bool StaticAllocator<T,S,I>::_initialized;


template<class T, int S, int I>
typename StaticAllocator<T,S,I>::pointer_t StaticAllocator<T,S,I>::allocate(){
	if (_pos < StaticAllocator<T,S,I>::SIZE){
		T* ret = _ptr[_pos];
		++_pos;
		return static_cast<Ptr>(ret);
	}
	return 0;		
}

// fuer Chain<T> spezialisierte Variante, die die Struktur der zu verwaltenden 
// Objekte fuer sich selbst ausnutzt. (bringt nur Vorteile bei AVR und H8)

//! Specialized \ref StaticAllocator for \ref Chain 
/**
 * This specialized implementation uses the Structure of the \ref Chain class 
 * in order to reduce the memory usage. 
 */
template<class T,int S, int I>
class StaticAllocator< Chain<T>,S,I> { 
  
// public typedefs and enums
public:
	
	//! Type definition for instatiated type
	typedef Chain<T> value_t;
	//! Type definition for pointer to intantiated type
	typedef Ptr pointer_t;
	//! Type definition for this type of allocator
	
	
	typedef typename index_t_value<S>::T index_t;
	
	typedef StaticAllocator< Chain< T > , S > alloc_t;
  	   
  	enum {SIZE=S};
  	   
  	
// protected variables   
protected:
	//! pointer to the next free memory location
	static value_t* _head;
	//! memory area for storing objects of \ref value_t
	static UInt8 _pool[ S * sizeof(value_t) ];

public:
	
	//! constructor
	/**
	 * Initialize the structure of the memory pool by linking together the 
	 * \ref Chain objects.
	 */
	StaticAllocator(){
		_head = (value_t*) _pool;
		for(int i=0;i<SIZE-1;++i){
			//pool(i)->setNext(pool(i+1));
			((value_t*) &_pool[i*sizeof(value_t)])->setNext((value_t*) &_pool[(i+1)*sizeof(value_t)]);
		}
	}
	
	//! TODO
	StaticAllocator(const StaticAllocator& alloc)
	{}
	
	//! TODO
	pointer_t allocate();
	
	//! TODO
	pointer_t allocate(size_t size);
	
	//! TODO
	void free(pointer_t elem);
};


template<class T, int S,int I>
void* StaticAllocator<Chain<T>,S,I>::allocate(){
		Chain<T>* ret=_head;
		if(_head != 0){
			_head=(Chain<T>*) _head->getNext();
		}
		return static_cast<pointer_t>(ret);
}

template<class T, int S,int I>
void* StaticAllocator<Chain<T>,S,I>::allocate(size_t size){
		return allocate();
}

template<class T, int S,int I>
void StaticAllocator<Chain<T>,S,I>::free(pointer_t elem){
		(static_cast< Chain<T>* >(elem))->setNext(_head);
		_head=(static_cast< Chain<T>* >(elem));
}

template<class T, int S,int I>	
UInt8 StaticAllocator<Chain<T>,S,I>::_pool[ S * sizeof(Chain<T>) ];

template<class T, int S,int I>	
Chain<T>* StaticAllocator<Chain<T>,S,I>::_head;


// Below there's some code which initializing the _ptr array at compile time.
// Unfortunatly the g++ compiler does not honor this approach with a smaller
// code size... 

//template<class A,int N>
//struct cont :public cont<A,N-1>{
//        static UInt8* data;
//};
//
//template<class A,int N>
//UInt8* cont<A,N>::data=(UInt8*) (cont<A,N-1>::data + sizeof(typename A::value_t));
//
//template<class A>
//struct cont<A,0>{
//        static  UInt8* data;
//};

//template<class A>
//UInt8*  cont<A,0>::data=A::_pool;

//template<class T, int S, int ID>	
//typename  index_t_value<S>::T StaticAllocator<T,S,ID>::_pos=S;
	
//template<class T, int S, int ID>	
//T** StaticAllocator<T,S,ID>::_ptr= (T**) &cont< StaticAllocator<T,S,ID> , S >::data ;


#endif //__StaticAllocator_h__
