package SystemConfiguration;

use warnings;
use strict;

use Carp;
use XML::Simple;
use Data::Dumper;

sub readOSConfig {
	my $osvariant = shift;
	my %osvariant = ( );
	my $line = 0;

	open(OSCFG, $osvariant) || croak "Cannot open $osvariant\n";
	while(<OSCFG>) {
		$line++;
		chomp;
		next if /^\s*(#.*)?$/;
		unless( /^CONFIG_([^=]+)=(.*)$/ ) {
			carp "parse error: $osvariant:$line: $_\n";
			next;
		}
		$osvariant{$1} = $2;
	}
	close(OSCFG);
	return \%osvariant;
}

sub readAppDef {
	my $appdefinition = shift;

	my $xml = XMLin($appdefinition,
		forcearray => [ qw(alarm alarmref applicationheader applicationmode component componentref counter counterref event eventref hooksheader isr isrref messageport networkport networkportref osapplication receiverport resource senderport sharedmemory shmref task taskref) ],
		GroupTags => {
			'autostarttasks' => 'taskref',
			'autostartalarms' => 'alarmref',
			'osapplication' => 'componentref',
			keyattr => ['name'],
			ValueAttr => [ qw(componentref isrref networkportref receiverport senderport shmref taskref) ]
		},
	) || croak "could not read $appdefinition\n";

}

# preprocess the hash generated from the xml-file
sub processAppDef {
	my ($os, $appcfg, $ap) = @_;

	foreach my $appk (keys %{$appcfg->{osapplication}}) {
		my $app = $appcfg->{osapplication}->{$appk};
		SystemConfiguration::OSApplication->init($app, $appk);

		# set back references in components
		foreach my $ck (keys %{$app->{componentref}}) {
			my $c = $appcfg->{component}->{$ck};
			$c->{osapp} = $app;
			$app->{componentref}->{$ck} = $appcfg->{component}->{$ck};

			# set task references within components
			foreach my $taskk (keys %{$c->{taskref}}) {
				my $taskh = $appcfg->{task}->{$taskk};
				$c->{taskref}->{$taskk} = $taskh;
				$taskh->{component} = $c;
			}

			# set ISR references within components
			foreach my $isrk (keys %{$c->{isrref}}) {
				my $isrh = $appcfg->{isr}->{$isrk};
				$c->{isrref}->{$isrk} = $isrh;
				$isrh->{component} = $c;
			}
		}
	}

	foreach my $mpk (keys %{$appcfg->{messageport}}) {
		my $mp = $appcfg->{messageport}->{$mpk};
		SystemConfiguration::MessagePort->init($mp, $mpk);
	}

	# Create networkport objects and cross reference the messageport objects
	foreach my $netpk (keys %{$appcfg->{networkport}}) {
		my $netp = $appcfg->{networkport}->{$netpk};
		$netp->{sendercomponent} = $appcfg->{messageport}->{$netp->{senderport}};
		$netp->{receivercomponent} = $appcfg->{messageport}->{$netp->{receiverport}};
		SystemConfiguration::NetworkPort->init($netp, $netpk);
	}

	foreach my $shmk (keys %{$appcfg->{sharedmemory}}) {
		my $shm = $appcfg->{sharedmemory}->{$shmk};
		SystemConfiguration::SharedMemory->init($shm, $shmk);
	}

	foreach my $ck (keys %{$appcfg->{component}}) {
		my $c = $appcfg->{component}->{$ck};
		$c->{name} = $ck;

		# set sender component references in message ports
		foreach my $mpk ( keys %{$c->{senderport}} ) {
			my $mp = $appcfg->{messageport}->{$mpk};
			$mp->{sendercomponent} = $c;
			$ap->{senderPortsWithWriteRevocation} = 1 if exists $c->{senderport}->{$mpk}->{revokewriteaccess} && $c->{senderport}->{$mpk}->{revokewriteaccess} eq 'true';
		}

		# set receiver component references in message ports
		foreach my $mpk ( keys %{$c->{receiverport}} ) {
			my $mp = $appcfg->{messageport}->{$mpk};
			$mp->{receivercomponent} = $c;
			$ap->{receiverPortWithWriteAccess} = 1 if exists $c->{receiverport}->{$mpk}->{grantwriteaccess} && $c->{receiverport}->{$mpk}->{grantwriteaccess} eq 'true';
		}

		# register shared memory access
		foreach my $shmk ( keys %{$c->{shmref}} ) {
			my $shm = $appcfg->{sharedmemory}->{$shmk};
			die "shared memory $shmk referenced but not defined\n" unless ref $shm;
			my $mode = $c->{shmref}->{$shmk}->{mode};
			$shm->registerReader($c) if $mode =~ /r/;
			$shm->registerWriter($c) if $mode =~ /w/;
		}
	}

	# bless tasks
	foreach my $taskk (keys %{$appcfg->{task}}) {
		my $taskh = $appcfg->{task}->{$taskk};
		SystemConfiguration::Task->init($taskh, $taskk);
	}

	# bless isrs
	foreach my $isrk (keys %{$appcfg->{isr}}) {
		my $isrh = $appcfg->{isr}->{$isrk};
		SystemConfiguration::ISR->init($isrh, $isrk);
	}

	# bless counters
	foreach my $ck (keys %{$appcfg->{'counter'}}) {
		my $ch = $appcfg->{'counter'}->{$ck};
		SystemConfiguration::Counter->init($ch, $ck);
	}

	# bless alarms
	foreach my $ak (keys %{$appcfg->{'alarm'}}) {
		my $ah = $appcfg->{'alarm'}->{$ak};
		SystemConfiguration::Alarm->init($ah, $ak, $appcfg);
	}

	foreach my $ak (keys %{$appcfg->{'resource'}}) {
		my $ah = $appcfg->{'resource'}->{$ak};
		$ah->{name} = $ak;
	}

	# appmodes (must be last)
	foreach my $amk (keys %{$appcfg->{applicationmode}}) {
		my $am = $appcfg->{applicationmode}->{$amk};
		SystemConfiguration::AppMode->init($am, $amk, $appcfg);
	}
}

# check consistency of app and OS configuration and print warnings
sub consistencyChecks {
	my ($os, $appcfg, $ap) = @_;

}

sub hasOSFeature {
	my ($self, $ft) = @_;

	return $self->{os}->{$ft} if exists $self->{os}->{$ft};
	return undef;
}

sub getMessagePorts {
	my $self = shift;
	return values %{$self->{app}->{messageport}};
}

sub getSharedMemories {
	my $self = shift;
	return values %{$self->{app}->{sharedmemory}};
}

sub getEvent {
	my $self = shift;
	my $name = shift;

	return $self->{app}->{event}->{$name}
}

sub getEvents {
	my $self = shift;
	return keys %{$self->{app}->{event}};
}

sub getResources {
	my $self = shift;
	return values %{$self->{app}->{resource}};
}

sub getAlarms {
	my $self = shift;
	return values %{$self->{app}->{'alarm'}};
}

sub getCounters {
	my $self = shift;
	return values %{$self->{app}->{'counter'}};
}

sub getCounter {
	my ($self,$name) = @_;
	unless ( exists $self->{app}->{'counter'}->{$name} ) {
		print STDERR "Warning: Request for unknown counter $name\n";
		return undef;
	}
	return $self->{app}->{'counter'}->{$name};
}

sub getAppModes {
	my $self = shift;
	return values %{$self->{app}->{applicationmode}};
}

sub getApplications {
	my $self = shift;
	return sort {
		if($a->{trusted} eq 'true' && $b->{trusted} eq 'true') {
			return 0;
		}
		if($a->{trusted} eq 'true') {
			return -1;
		}
		if($b->{trusted} eq 'true') {
			return 1;
		}
		$a->{name} cmp $b->{name}
	} values %{$self->{app}->{osapplication}};
}

sub getComponents {
	my $self = shift;
	return values %{$self->{app}->{component}};
}

sub getTasks {
	my $self = shift;
	return sort { 
		$a->{priority} <=> $b->{priority} ||
		$a->{name} cmp $b->{name}
	} values %{$self->{app}->{task}};
}

sub getISRs {
	my $self = shift;
	return values %{$self->{app}->{isr}};
}

sub getAppHeaders {
	my $self = shift;
	return keys %{$self->{app}->{applicationheaders}->{applicationheader}};
}

sub getHookHeaders {
	my $self = shift;
	return keys %{$self->{app}->{hooksheaders}->{hooksheader}};
}

sub hookFunction {
	my ($self, $hook) = @_;
	return undef unless exists $self->{app}->{$hook};
	return $self->{app}->{$hook}->{name};
}

sub getNetworkPorts {
	my $self = shift;
	return values %{$self->{app}->{networkport}};
}

sub new {
	my ($class, $appdefinition, $osvariant) = @_;
	my $self = { 
		os  => readOSConfig($osvariant),
		app => readAppDef($appdefinition),
		appprops => { },
	};

	processAppDef($self->{os}, $self->{app}, $self->{appprops});

	consistencyChecks($self->{os}, $self->{app}, $self->{appprops});

	return bless $self, $class;
}

package SystemConfiguration::OSApplication;

sub init {
	my ($class, $self, $name) = @_;
	$self->{name} = $name;
	bless $self, $class;
}

sub appSymbolPrefix {
	my $self = shift;
	return '__CIAO_APPDATA_'.$self->{name}.'_';
}

sub getName {
	$_[0]->{name};
}

package SystemConfiguration::Task;

sub init {
	my ($class, $self, $name) = @_;
	$self->{name} = $name;
	bless $self, $class;
}

sub entryFunction {
	my $self = shift;
	return $self->{component}->{name}.'::functionTask'.$self->{name};
}

sub priority {
	return $_[0]->{priority};
}

package SystemConfiguration::Alarm;

sub init {
	my ($class, $self, $name, $appcfg) = @_;
	$self->{name} = $name;
	for my $attr qw(absolutetime cycletime) {
		$self->{$attr} = 0 unless exists $self->{$attr};
	}

	if(exists $self->{'counterref'}) {
		$self->{counter} = $appcfg->{counter}->{$self->{'counterref'}};
	} else {
		die "Alarm $name lacks the required attribute counterref\n";
	}
	bless $self, $class;
}

sub getCounter {
	my $self = shift;
	return $self->{counter};
}

package SystemConfiguration::Counter;

sub init {
	my ($class, $self, $name) = @_;
	$self->{name} = $name;
	for my $attr qw(maxallowedvalue mincycle ticksperbase) {
		die "Counter $name lacks the required attribute $attr\n" unless exists $self->{$attr};
	}
	bless $self, $class;
}

package SystemConfiguration::ISR;

sub init {
	my ($class, $self, $name) = @_;
	$self->{name} = $name;
	bless $self, $class;
}

sub entryFunction {
	my $self = shift;
	return $self->{component}->{name}.'::functionISR'.$self->{name};
}

sub interruptSource {
	my $self = shift;
	$self->{device};
}

sub getApplication {
	my $self = shift;
	$self->{component}->{osapp};
}

sub category {
	return $_[0]->{category};
}

package SystemConfiguration::SharedMemory;

sub init {
	my ($class, $self, $name) = @_;
	$self->{name} = $name;
	$self->{readers} = [];
	$self->{writers} = [];
	bless $self, $class;
}

sub registerReader {
	my ($self, $component) = @_;
	push @{$self->{readers}}, $component;
}

sub registerWriter {
	my ($self, $component) = @_;
	push @{$self->{writers}}, $component;
}

sub shmSymbol {
	my $self = shift;
	my $ps = $self->{name}.'__area';

	my @writers = @{$self->{writers}};
	if(@writers == 0) {
		print STDERR "Shared Memory area $self->{name} has no writers\n";
	} elsif(@writers == 1) {
		$ps = $writers[0]->{osapp}->appSymbolPrefix().$ps;
	} else {
		die "Multiple Writers are currently not supported for shared memory ($self->{name})";
	}
	$ps;
}

sub areaType {
	my $self = shift;
	$self->{ctype};
}

sub getWriters {
	my $self = shift;
	$self->{writers};
}

package SystemConfiguration::MessagePort;

sub init {
	my ($class, $self, $name) = @_;
	$self->{name} = $name;
	bless $self, $class;
}

sub poolType {
	my $self = shift;
	$self->{msgctype};
}

sub poolSymbol {
	my $self = shift;
	my $ps = $self->{name}.'__pool';
	unless (exists $self->{sendercomponent}->{senderport}->{$self->{name}}->{revokewriteaccess} &&
		$self->{sendercomponent}->{senderport}->{$self->{name}}->{revokewriteaccess} eq 'true') {
		$ps = $self->{sendercomponent}->{osapp}->appSymbolPrefix().$ps;
	}
	$ps;
}

sub poolSize {
	my $self = shift;
	$self->{msgcount};
}

sub senderPortType {
	my $self = shift;
	if(exists $self->{sendercomponent}->{senderport}->{$self->{name}}->{revokewriteaccess} &&
		$self->{sendercomponent}->{senderport}->{$self->{name}}->{revokewriteaccess} eq 'true') {
		return 'os::mp::SenderPortWriteProtection';
	}
	return 'os::krn::SenderPort';
}

sub senderPortSymbol {
	my $self = shift;
	$self->{sendercomponent}->{osapp}->appSymbolPrefix().$self->{name}.'__sender';
}

sub senderPortInit {
	my $self = shift;
	my $rp = $self->receiverPortSymbol();
	if( exists $self->{receivercomponent}->{receiverport}->{$self->{name}}->{grantwriteaccess} &&
		$self->{receivercomponent}->{receiverport}->{$self->{name}}->{grantwriteaccess} eq 'true') {
		$rp.= '.rp';
	}
	my $i = "&$rp, 0, ".$self->poolSymbol();
	if(exists $self->{sendercomponent}->{senderport}->{$self->{name}}->{revokewriteaccess} &&
		$self->{sendercomponent}->{senderport}->{$self->{name}}->{revokewriteaccess} eq 'true') {
		$i = "{ $i }";
	}
	return $i;
}

sub receiverPortType {
	my $self = shift;
	if( exists $self->{receivercomponent}->{receiverport}->{$self->{name}}->{grantwriteaccess} &&
		$self->{receivercomponent}->{receiverport}->{$self->{name}}->{grantwriteaccess} eq 'true') {
		return 'os::mp::ReceiverPortWriteAccess';
	}
	return 'os::krn::ReceiverPort';
}

sub receiverPortSymbol {
	my $self = shift;
	$self->{receivercomponent}->{osapp}->appSymbolPrefix().$self->{name}.'__receiver';
}

sub receiverPortInit {
	my $self = shift;
	my $sp = $self->senderPortSymbol();
	if(exists $self->{sendercomponent}->{senderport}->{$self->{name}}->{revokewriteaccess} &&
		$self->{sendercomponent}->{senderport}->{$self->{name}}->{revokewriteaccess} eq 'true') {
		$sp.= '.sp';
	}
	my $i = "&$sp, 0";
	if( exists $self->{receivercomponent}->{receiverport}->{$self->{name}}->{grantwriteaccess} &&
		$self->{receivercomponent}->{receiverport}->{$self->{name}}->{grantwriteaccess} eq 'true') {
		$i = "{ $i }";
	}
	return $i;
}

sub templateParams {
	my $self = shift;
	"<".$self->poolType().", ".$self->poolSize().">";
}

package SystemConfiguration::AppMode;

sub init {
	my ($class, $self, $name, $appcfg) = @_;
	$self->{name} = $name;
	foreach my $tk (keys %{$self->{autostarttasks}}) {
		$self->{autostarttasks}->{$tk} = $appcfg->{task}->{$tk};
	}
	foreach my $ak (keys %{$self->{autostartalarms}}) {
		$self->{autostartalarms}->{$ak} = $appcfg->{'alarm'}->{$ak};
	}
	bless $self, $class;
}

sub getTasks {
	my $self = shift;
	return values %{$self->{autostarttasks}};
}

sub getAlarms {
	my $self = shift;
	return values %{$self->{autostartalarms}};
}


package SystemConfiguration::NetworkPort;

sub init {
	my ($class, $self, $name) = @_;
	$self->{name} = $name;
	bless $self, $class;
}

sub senderComponent {
	my $self = shift();
	return $self->{sendercomponent}
}

sub port {
	my $self = shift();
	return $self->{port};
}

sub task {
	my $self = shift();
	return $self->{taskref};
}

1;
