#!/usr/bin/perl

use strict;
use warnings;
use Hash::Util;

my %sections = (
	'b' => 'bss',
	'd' => 'data',
	't' => 'text',
);

my %symbols = ( );
$symbols{$_} = { 
	attr => { },
	symbols => { },
} foreach(values %sections);

# {
#		'section' => {
#			'symbols' => {
#				'symbol' => {
#					'rtcs' => {
#						'checkNameLong/checkAbbrv' => 3,
#					},
#					variants => {
#						'variant' => {
#							'size' => 23,
#							'addr' => 0x349834,
#							'global' => (0|1),
#						},
#					},
#				},
#			},
#			'attr' => {
#				'variant' => {
#					size => 23,
#				},
#			},
#		},
# }

our @variants = ();
our $verbose=0;
our $showAll=0;
my  $rtcsFile;

foreach (@ARGV) {
	if(/^--/) {
		if($' eq 'verbose') {
			$verbose++;
		} elsif($' eq 'showAll') {
			$showAll=1;
		} elsif($' =~ /rtcs=/) {
			$rtcsFile = $';
		}
	} else { # variant
		my ($v,$e) = ($_,$_);
		($v,$e) = ($`,$') if /:/;
		push @variants, {
			variant => $v,
			elffile => $e,
		};
	}
}

if(@variants==0) {
	print STDERR "Usage: $0 [<switches>] [Variantname:]Reference-Elffile [[Variantname:]Elffile ...]\n";
	print STDERR "Available switches:\n";
	print STDERR "--rtcs=<path>  file containing runtime check numbers for text symbols\n";
	print STDERR "--showAll      output information on all symbols (default: only symbols where sizes in variants differ)\n";
	print STDERR "--verbose      increase level of verbosity\n";
	exit 1;
}

# first variant is reference
our $refVariant = $variants[0]->{variant};

# read symbols and sizes of each variant
foreach my $var (@variants) {
	my	$variant = $var->{variant}; 
	my	$elffile = $var->{elffile};

	open(NM, "tricore-nm -CS $elffile|") || die "cannot open $elffile: $@\n";

	while(my $line = <NM>) {
		chomp($line);
		next unless($line =~ /([0-9a-fA-F]{8}) ([0-9a-fA-F]{8}) ([A-Za-z]) ([^ ]+)/);

		my ($addr, $ssize, $sect, $symname) = ($1, $2, $3, $4);
		$symname =~ s/\(.*//;

		unless(exists $sections{lc($sect)}) {
			print STDERR "skipping unknown section $sect ($symname)\n" if $verbose>0;
			next;
		}

		my $section = $sections{lc($sect)};
		my $sym = {
			addr   => hex($addr),
			size   => hex($ssize),
			global => (lc($sect) ne $sect),
		};

		$symbols{$section}->{symbols}->{$symname} = { variants => { } } unless exists $symbols{$section}->{symbols}->{$symname};
		$symbols{$section}->{symbols}->{$symname}->{variants}->{$variant} = $sym;

		# section attributes
		$symbols{$section}->{attr}->{$variant} = { size=>0 } unless exists $symbols{$section}->{attr}->{$variant};
		$symbols{$section}->{attr}->{$variant}->{size} += $sym->{size};
	}

	close(NM);
}

if($rtcsFile) {
	open(RTCS, $rtcsFile) || die "Failed to open runtime check stat file $rtcsFile\n";
	# First line contains check descriptions
	my @checkdesc = split(/ /, <RTCS>);
	die "Invalid rtCheckStatFile format\n" unless shift @checkdesc eq 'symbol';

	# All other lines contain the check numbers for a text symbol
	while(<RTCS>) {
		chomp;
		my ($sym, @checkNumbers) = split(/ /);
		die "Number of checks does not match number of check descriptions\n" unless @checkNumbers == @checkdesc;
		my $tsyms = $symbols{$sections{t}}->{symbols};
		next unless exists $tsyms->{$sym};
		my %rtcsh = map { $checkdesc[$_] => $checkNumbers[$_] } 0 .. $#checkNumbers;
		$tsyms->{$sym}->{rtcs} = \%rtcsh;
	}

	close(RTCS);
}

foreach my $section (sort keys %symbols) {
	print "****************************************************************************\n";
	print "                              Section $section\n";
	print "****************************************************************************\n";
	
	print "Size (bytes):\n";
	printSizes($symbols{$section}->{attr}, "\t");

	print "\nSymbolsizes (bytes):\n";
	foreach my $sym (sort keys %{$symbols{$section}->{symbols}}) {
		my $symh = $symbols{$section}->{symbols}->{$sym};
		next unless $showAll || varSizesDiffer($symh);

		print "\t$sym\n";
		printSizes($symh->{variants}, "\t\t");
		if(exists $symh->{rtcs}) {
			print "\t\truntime check stats:\n";
			printrtCheckStat($symh->{rtcs});
		}
	}
}

sub printSizes {
	my $href = shift;
	my $prefix = shift;
	my $showPercent = 0;

	foreach my $vi (0 .. $#variants) {
		my $var = $variants[$vi]->{variant};
		print "$prefix$var: ";

		if(exists $href->{$var}) {
			my $vsize = $href->{$var}->{size};
			print $vsize;

			if($showPercent) {
				my $rsize = $href->{$refVariant}->{size};
				printf " (%+.2f%%)", ($vsize-$rsize)/$rsize*100;
			}

			$showPercent = 1 if $vi==0;
		} else {
			print "n/a";
		}
		print "\n";
	}
}

sub printrtCheckStat {
	my $rtcs = shift;

	foreach(sort keys %$rtcs) {
		my ($s,$l) = split('/');
		printf "\t\t\t%3d %s\n", $rtcs->{$_}, $l if $rtcs->{$_}>0;
	}
}

sub varSizesDiffer {
	my $symh = shift;

	return 1 unless exists $symh->{variants}->{$refVariant};
	my $rsize = $symh->{variants}->{$refVariant}->{size};

	foreach (keys %{$symh->{variants}}) {
		return 1 if $rsize != $symh->{variants}->{$_}->{size};
	}
	return 0;
}

