/*****************************************************************************
 * CiAO Proof-of-Concept: Protection
 * 
 * This sample program shows, how memory protection can be achieved in a 
 * statically linked system. The idea is to use a linker script (file: sections)
 * that places the os layer in one section and the ciao layer in another section. 
 * The address range of the os section is protected by the MPU. To enter the
 * kernel, a trap has to be performed, which is inserted by around advice into the
 * calls to os::XXX from ciao::XXX. The action object (to be located on the 
 * still accessable user stack) is passed to the trap handler and, after 
 * validation, used to trigger the real os::XXX call. 
 * 
 * Validation checks if the
 * call comes from the ciao section, as otherwise the trap handler has
 * been invoked directly, which is unsafe. This requires the ciao-layer to be
 * not inlined (cfCIAO_INLINE must not be defined). The os-layer, however, can
 * be inlined (cfOS_INLINE may be defined).
 * 
 * As the privileged call is routed back to the ciao layer (the generated action
 * object's trigger method is located there), the code from this layer has to be 
 * readable/executable but must not be writable for the application. The os layer 
 * should be protected completly. 
 *   
 *****************************************************************************/



#include <stdio.h>
#include "ciao/StdSerial.h"

#ifdef cfCIAO_TRAP_OSCALLS
	#include "ostrap.ah"
#endif

extern "C" int __OS_BEGIN;		// begin of the os section
extern "C" int __OS_END;
extern "C" int __CIAO_BEGIN;	// begin of the ciao section
extern "C" int __CIAO_END;

#ifdef cfCIAO_TRAP_OSCALLS

void bad_code() {
	
	printf("(-:  I am the bad code that should _never_ be callable via a trap!  :-)\n");
}


// Even if in the ciao namespace, this code will be considered "bad" as 
// the linker script additionally filters on the source location (ciao folder)
namespace ciao { 
	void bad_code_action(AC::Action&) {
		bad_code();
	}
	
	__attribute__((noinline))
	void bad_code_invoker(const char* arg)  {
		// try to setup an own action object, to run own code in privileged mode
		AC::Action a;
		a._wrapper = bad_code_action;
		
		ostrap::trap( a );	
	}	
}

#endif


int main(){
	using namespace ciao;
	
	printf("main()\n");
	printf("  __OS_BEGIN  = 0x%p\n", &__OS_BEGIN );
	printf("  __OS_END    = 0x%p\n", &__OS_END );
	printf("  __CIAO_BEGIN= 0x%p\n", &__CIAO_BEGIN );
	printf("  __CIAO_END  = 0x%p\n", &__CIAO_END );
	
	// this will "trap" into the kernel
	StdSerial::send("Hallo von main :-)");

#ifdef cfCIAO_TRAP_OSCALLS
 	// try to "insert" some code into the kernel
	bad_code_invoker("Hallo von main :-)");
#endif
	
}
